# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

The reference maps which map from a subject specific reference space to
the location and orientation of the subject's brain in the scanner
during the course of a FMRI session at a given time.

"""

from .tau import tau

from .affines import Affine, Affines

from .grubbs import grubbs

from .tracking import fit_by_pca

from .euler import rotation_matrix_to_euler_angles

import numpy as np

from numpy.linalg import inv, norm

import pickle

class ReferenceMaps:
    """
    The reference maps which map from a subject specific reference space
    to the location and orientation of the subject's brain in the
    scanner during the course of a FMRI session at a given time.

    Parameters
    ----------
    name : Identifier
    """

    def __init__(self, name):
        self.name = name

    def fit(self, session, use_raw=True):
        """
        Fit head movement

        This will fit rigid body transformations to the data that estimate
        position and bearing of the head.

        Parameters
        ----------
        session : Session
        use_raw : bool

        Notes
        -----
        The function implements the principle axis method for rigid body
        tracking.
        """
        self.shape = (session.numob, session.shape[session.ep])
        self.epi_code = session.epi_code
        self.ep = abs(self.epi_code)-1

        self.temporal_resolution = session.temporal_resolution
        self.slice_time = session.slice_time

        if use_raw:
            scan_references, w = fit_by_pca(
                    data=session.raw,
                    reference=session.reference)
        else:
            scan_references, w = fit_by_pca(
                    data=session.data,
                    reference=session.reference)

        self.semi_axis_norms = w

        self.set_scan_references(scan_references=scan_references)

    ####################################################################
    # Flights between space and time
    ####################################################################

    def set_scan_references(self, scan_references):
        """
        Sets the reference space of the fMRI experiment

        This will define the reference space of an fMRI experiment and
        calculates the corresponding observation grid of to this space.

        Parameters
        ----------
        scan_references : Affines or ndarray, shape (n,4,4), dtype: float
            The rigid body transformations.  Affine transformations that
            map coordinates, given with respect to the rigid body, to
            the coordinates of these points in (physical) scanner space
            at the time of the respected scan.

        Notes
        -----
        In particular, this function will set the scan references and
        their respected inverses.  In fact, the reference space is
        solely defined through these affine transformations.  A scan
        reference is a rigid body transformation that maps a point in
        the reference space `R` of a rigid body to the position of this
        point in the scanner `V` during the respected scan, in signs:

        :math:` R \to V`

        Note that the map goes from *reference* to *specific*, or from
        *the* scan to *some* scan, or from *fixed* to *moving*.

        The reference space is typically set by a function that fits
        head movements to observed intensities at a fixed (say, scanner
        specific) grid, e.g., `fit_head_movements`.
        """
        if type(scan_references) is Affines:
            self.scan_references = scan_references
        else:
            self.scan_references = Affines(scan_references)

    def reset_reference_space(self, x=None):
        """
        This will reset the reference space

        Parameters
        ----------
        x : None or Affine or ndarray, shape (4,4), dtype: float
            an affine transformation or None (default)

        Notes
        -----
        This will reset the reference space by moving origin and base
        vectors to the new position specified by the affine
        transformation x.  The affine transformation goes **from this**
        reference space **to the new** reference. If the affine
        transformation has the form Ax+b, then b is the new origin and A
        defines the new orienation.

        If x=None, then x will be set to the mean of the scan
        references.

        As scan references :math:`ρ_t` go from scanner space to
        reference space, the new scan reference are

        .. math::

            ρ_t ∘ x^{-1}

        A warning: this makes all population maps which have been
        defined for this session obsolete. You should run this
        function before setting any population maps!
        """
        if x is None:
            x = self.scan_references.mean_rigid()

        if type(x) is not Affine:
            x = Affine(x)

        assert type(x) is Affine, 'x must be Affine'

        self.reference_affine = x
        scan_references  = x.inv().dot(self.scan_references)
        self.set_scan_references(scan_references=scan_references)

    ####################################################################
    # Outlier detection
    ####################################################################

    def detect_outlying(self, sgnf):
        """
        Detect outlying scans

        Parameters
        ----------
        sgnf : float
            Significance level at which a test for the existence of a
            outlier is tested.

        Returns
        -------
        ndarray of bool
             True if the particular full scan cycle is considered an outlier.

        Notes
        -----
        Uses the eigenvalues of the pcm method to for outlier detection.
        (Currently only full scan cycles supported.)

        Remove scan cycles which have eigenvalues which differ
        significantly from all other volumes, as these likely are the
        result of severe head movement during this particular
        measurement period.

        Grubbs' test is used recursively for the outlier detection. The
        norm of all three semi axis length and each single semi axis
        length is tested for outlier separately.

        It should simply give you an idea on how severe head movements
        might have been, and allow you to remove the most obvious
        outliers.
        """
        # TODO:
        # 1. Implement 'less' in `grubbs_test` and change outlying
        #    cycle detection to less.
        # 2. Do this with respect to each slice not with respect to a
        #    full scan cycle...
        # 3. Use as the reference distribution for grubbs only scans
        #    within blocks of irritation

        # Look at the length of the semi axis norms
        w0 = self.semi_axis_norms[...,0]
        w1 = self.semi_axis_norms[...,1]
        w2 = self.semi_axis_norms[...,2]
        _, args0 = grubbs(w0, sgnf)
        _, args1 = grubbs(w1, sgnf)
        _, args2 = grubbs(w2, sgnf)

        # Look at the bary centres of the scan cycles
        x0 = self.scan_references.affines[:,0,3]
        x1 = self.scan_references.affines[:,1,3]
        x2 = self.scan_references.affines[:,2,3]
        _, args3 = grubbs(x0, sgnf)
        _, args4 = grubbs(x1, sgnf)
        _, args5 = grubbs(x2, sgnf)

        euler = self.scan_references.euler().T

        euler[ abs(euler) > tau / 5 ] = np.nan

        _, args6 = grubbs(euler[0], sgnf)
        _, args7 = grubbs(euler[1], sgnf)
        _, args8 = grubbs(euler[2], sgnf)

        values = np.vstack((
            w0, w1, w2,
            x0, x1, x2,
            euler[0], euler[1], euler[2]))

        outlying = np.vstack((
            args0, args1, args2,
            args3, args4, args5,
            args6, args7, args8))

        self.outlying = outlying
        self.values   = values

    def detect_outlying_scans(self, sgnf):
        self.detect_outlying(sgnf)

        self.outlying_cycles = self.outlying.any(axis=0)
        self.outlying_scans = np.repeat(
                self.outlying_cycles,
                self.shape[-1]).reshape(self.shape)

    ####################################################################
    # Descriptive statistics
    ####################################################################

    def descriptive_statistics(self, sgnf=.1):
        """
        Give descriptive statistics of the instance
        """
        self.detect_outlying_scans(sgnf=sgnf)
        n, m = self.shape
        o = self.outlying_cycles.sum()
        return n, m*n, o, m*o

    def describe(self):
        """
        Give a description of the instance
        """
        description = """
        Scans:        {:>4d}
        Valid:        {:>4d}
        Outlying:     {:>4d} ({:.2f}%)
        Scan cycles:  {:>4d}
        Valid:        {:>4d}
        Outlying:     {:>4d} ({:.2f}%)
        """
        n, m, on, om = self.descriptive_statistics()

        return description.format(
                m, m-om, om, 100*(om/m),
                n, n-on, on, 100*(on/n))

    ####################################################################
    # Save and load class instance to and from disk
    ####################################################################

    def save(self, file, **kwargs):
        """
        Save instance to disk

        This will save the instance to disk for later use.

        Parameters
        ----------
        file : str
            File name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Wrapper for FSL

"""

from .pmap import PopulationMap

from .diffeomorphisms import Warp

from .nifti import nii2image, image2nii

import nibabel as ni

import numpy as np

from subprocess import PIPE

from subprocess import run

import os

from os.path import isfile, isdir, join

def bet(intercept, intercept_file, mask_file, cmd, variante='R',
        verbose=0):

    dfile = os.path.dirname(intercept_file)
    if dfile and not isdir(dfile):
       os.makedirs(dfile)

    ni.save(image2nii(intercept), intercept_file)

    command = [cmd]
    command.append(intercept_file)
    command.append(mask_file)
    command.append('-{}'.format(variante))

    if verbose:
        print('\n  '.join(command))

    command = ' '.join(command)

    try:
        bout = run([command], shell=True, stdout=PIPE, check=True)
        bout = bout.stdout.decode('utf-8').strip()
        if verbose > 1 and bout != '':
            print('Output of bet: {}'.format(bout))
    except Exception as e:
        print('Unable to run: {}'.format(command))
        print('Failed with: {}'.format(e))
        return

    try:
        mask = ni.load(mask_file)
    except Exception as e:
        print('Unable to read: {}'.format(mask_file))
        print('Failed with: {}'.format(e))
        return

    template = nii2image(mask)
    return template

def fit_warpcoef(nb_file, warpcoef_file, preimage_file=None,
        vb_file=None, vb_mask=None, nb_mask=None,
        config='T1_2_MNI152_2mm', cmd='fsl5.0-fnirt', verbose=True):
    """
    Run FNIRT to fit a diffeomorphism :math:`ψ` given data :math:`M` in
    the domain (vb) and data :math:`R` in the image (nb) of the
    diffeomorphism.

    Parameters
    ----------
    nb_file : str
        File name where to find the data of :math:`R=ψ[M]` which shall
        be used as reference for the image space (nb) of the
        diffeomorphism.
    warpcoef_file : str
        File name in which to save the warp coefficients.
    target_file : None or str
        File name in which to save the target :math:`ψ^{-1}[R]`. If
        None, the target will not be saved. This is the default.
    config : str
        Name of the configuration file FNIRT shall use.
    vb_file : None or str
        File name where to find the data of :math:`M` (the template)
        which shall be used as reference for the domain space (nb) of
        the diffeomorphism.
    vb_mask : None or str
        Name of file with mask in reference space. May be provided if
        vb_file is given.
    cmd : str
        Name of the FSL's FNIRT command line program.
    verbose : bool
        Print concatinated command to stdout
    """
    command = [cmd]

    command.append('--in={}'.format(nb_file))

    if nb_mask:
        command.append('--inmask={}'.format(nb_mask))

    if vb_file:
        command.append('--ref={}'.format(vb_file))

    if vb_mask:
        command.append('--refmask={}'.format(vb_mask))

    if config:
        command.append('--config={}'.format(config))

    if preimage_file:
        command.append('--iout={}'.format(preimage_file))

    command.append('--cout={}'.format(warpcoef_file))

    if verbose:
        print('\n  '.join(command))

    command = ' '.join(command)

    try:
        sout = run([command], shell=True, stdout=PIPE, check=True)
        if verbose > 1:
            print(sout.stdout.decode('utf-8').strip())
        return True
    except Exception as e:
        print('Unable to run: {}'.format(command))
        print('Failed with: {}'.format(e))
        return False

def warpcoef2pmap(warpcoef_file, vb_file, vb_name, nb_file, nb_name,
        cpopulation_file, csubject_file, cmd='fsl5.0-std2imgcoord'):
    """
    Run FSL's img2stdcoord to turn the warp coefficient file produced by
    FNIRT into a fmristats' diffeomorphism intance.

    Parameters
    ----------
    vb_file : str
        file name where to find the data of :math:`M` (the template)
        which shall be used as reference for the domain space (nb) of
        the diffeomorphism, i.e. file name in which the template is
        stored.
    nb_file : str
        file name where to find the data of :math:`R=ψ[M]` which shall
        be used as reference for the image space (nb) of the
        diffeomorphism.
    vb_name :
        name or identifyer for the domain (vb).
    nb_name :
        name or identifyer for the image (nb).
    warpcoef_file : str
        File name of the warp coefficient file.
    cmd : str
        Name of FSL's std2imgcoord command line program.
    cpopulation_file : str
        Coordinates of the template's image grid in population space
    csubject_file : str
        Coordinates of the template's image grid in subject space

    Returns
    -------
    ndarray, shape template.shape
        The coordinates.
    """
    template = nii2image(ni.load(vb_file))
    template_grid = template.coordinates()
    template_grid = template_grid.reshape(-1,3)
    np.savetxt(cpopulation_file, X=template_grid, delimiter=' ', fmt='%i')

    command = '{} -std {} -img {} -warp {} -mm {} > {}'.format(
            cmd, vb_file, nb_file, warpcoef_file,
            cpopulation_file, csubject_file)

    try:
        soutput = run([command], shell=True, stdout=PIPE, check=True)
    except Exception as e:
        print('Unable to run: {}'.format(command))
        print('Failed with: {}'.format(e))
        return

    try:
        coordinates = np.loadtxt(csubject_file)
    except Exception as e:
        print('Unable to read: {}'.format(csubject_file))
        print('Failed with: {}'.format(e))
        return

    coordinates = coordinates.reshape(template.shape+(3,))

    diffeomorphism = Warp(
            reference=template.reference,
            warp=coordinates,
            vb=vb_name,
            nb=nb_name,
            metadata={
                'vb_file':vb_file,
                'nb_file':nb_file,
                'warpcoef':warpcoef_file,
                }
            )

    return PopulationMap(diffeomorphism, template=template)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

Calculates the population effect field from fMRI data of the brain.

"""

import statsmodels.api as sm

import numpy as np

from numpy.linalg import solve, svd

def _kernel_matrix(h, d, x):
    _, s, v = svd(x.T)
    K = v[len(s):].T
    return K.dot(solve(K.T.dot(np.diag(h+d)).dot(K), K.T))

def _qprofile_analysis(h, y, d):
    """
    In case of no covariates, i.e., with only an intercept in the
    regression, we are in the case of an meta analysis. The q-profile
    funtion q_τ simplifies:

    .. math:

    q_δ(τ) = \sum_j \frac 1{δ_i + τ} ⋅ (y - \bar y)^2

    """
    return ((y - y.mean())**2 / (h+d)).sum()

def _qprofile_regression(h, y, d, x):
    return y.dot( _kernel_matrix(h=h,d=d,x=x) ).dot(y)

def hedge_estimator(y, d):
    """
    Hedge estimate for the heterogeneity

    Parameters
    ----------
    y : ndarray, shape (k,)
        The observations.
    d : ndarray, shape (k,)
        The heteroscedasticity vector.
    x : ndarray, shape (k,p)
        The covariates.

    Return
    ------
    λ : float
        The estimated heterogeneity.
    """
    n = len (y)
    return float( max(0, (((y - y.mean())**2).sum() - (d -  d/n).sum()) / (n-1)))

def hedge_type_estimator(y, d, x):
    """
    Hedge estimate for the heterogeneity

    Parameters
    ----------
    y : ndarray, shape (k,)
        The observations.
    d : ndarray, shape (k,)
        The heteroscedasticity vector.
    x : ndarray, shape (k,p)
        The covariates.

    Return
    ------
    λ : float
        The estimated heterogeneity.
    """
    H  = x.dot(solve(x.T.dot(x), x.T))
    E  = np.eye(x.shape[0]) - H
    resid_df = -np.diff(x.shape)
    return float( max(0, (y.dot(E).dot(y) - np.trace(E.dot(np.diag(d)))) / resid_df) )

def meta_analysis(y, d):
    """
    Random effect meta regression

    Will regress y onto x using a random effects meta regression model
    with heteroscedasticity d.  Heterogeneity will be estimated by Hedge.

    Parameters
    ----------
    y : ndarray, shape (k,)
        The observations.
    d : ndarray, shape (k,)
        The heteroscedasticity vector.

    Return
    ------
    β : ndarray, shape (p)
        The location parameters.
    t : ndarray, shape (p)
        Knapp-Hartung adjusted t-test statistic
    λ : float
        The heterogeneity.

    Notes
    -----
    In case of a meta analysis, many formula simplify.

    .. math:

        σ(\hat μ) = \frac 1 {\sum_j (1/ δ_j + τ)}
        \hat μ    = \frac {\sum_j (1/ δ_j + τ) * y_j}{\sum_j (1/ δ_j + τ)}
    """
    h   = hedge_estimator(y,d)
    w   = 1/(h+d)
    v   = 1 / w.sum()
    b   = (w * y).sum() / w.sum()
    rdf = len(y) - 1
    adj = _qprofile_analysis(h=h,y=y,d=d) / rdf
    t = b / np.sqrt(v * adj)
    return b, t, h, rdf

def meta_regression(y, d, x):
    """
    Random effect meta regression

    Will regress y onto x using a random effects meta regression model
    with heteroscedasticity d.  Heterogeneity will be estimated by Hedge.

    Parameters
    ----------
    y : ndarray, shape (k,)
        The observations.
    d : ndarray, shape (k,)
        The heteroscedasticity vector.
    x : ndarray, shape (k,p)
        The covariates.

    Return
    ------
    β : ndarray, shape (p)
        The location parameters.
    t : ndarray, shape (p)
        Knapp-Hartung adjusted t-test statistic
    λ : float
        The heterogeneity.
    """
    h   = hedge_type_estimator(y,d,x)
    fit = sm.WLS(y, x, weights=1/(h+d)).fit()
    v   = fit.bse
    b   = fit.params
    rdf = fit.df_resid
    adj = _qprofile_regression(h=h,y=y,d=d,x=x) / rdf
    t   = b / (v * np.sqrt(adj))
    return b, t, h, rdf

def fit_field(obs, design=None, mask=None):
    """
    Random effect meta regression

    Will regress each point of an effect and certainty field onto the
    covariates in x using a random effects meta regression model

    Parameters
    ----------
    obs : ndarray, shape 5D-image
        The observations.
    mask : ndarray, shape 3D-image
        A mask where to fit the field
    x : ndarray, shape (k,p)
        The covariates.

    Notes
    -----
    Results will be written into obs
    """
    if design is None:
        df = obs.shape[-1] - 1
        p  = 1
    else:
        df = -np.diff(design.shape)
        p  = design.shape[1]

    assert obs.shape[-1] > p+1, 'model not identifiable, too many parameters to fit'
    assert obs.shape[-1] > df, 'model not identifiable, too many parameters to fit'

    # pixels are 'fully valid' if there are no missing values along the
    # subject axis, i.e, for all subjects in the sample there is an
    # estimate available for this pixel.
    fully_valid = np.isfinite(obs).all(axis=(-1,-2))

    # pixels are 'valid' if there exists enough data such that the meta
    # regression / analysis model is identifiable.
    valid = np.isfinite(obs).all(axis=-2).sum(axis=-1)
    valid = valid > max(p+1,df)

    if mask is None:
        mask = valid
    else:
        assert mask.dtype is np.dtype(bool), 'mask must be of dtype bool or None'
        assert mask.shape == valid.shape, 'shape of mask does not fit'
        mask = mask & valid

    print("  … mask has shape: {}".format(mask.shape))

    assert mask.any(), 'mask is empty, i.e., there exists no valid pixels in this mask'

    # if the statement 'if valid, then fully_valid' holds, the following
    # loop don't need checks.  This is identical to 'not valid or
    # fully_valid'.
    skip_checks = (~mask | fully_valid).all()

    res = obs[mask]
    print("  … number of pixels to be fitted: {}".format(len(res)))

    it = iter(res)

    if skip_checks:
        print("  … all pixels in the mask are fully valid")
        if design is None:
            print("  … a meta analysis will be performed")
            for v in it:
                b, t, h, r = meta_analysis(y=v[0], d=v[1])
                v[0,0] = b
                v[1,0] = t
                v[0,1] = h
                v[1,1] = r
        else:
            print("  … a meta regression will be performed")
            for v in it:
                b, t, h, r = meta_regression(y=v[0], d=v[1], x=design)
                v[0,:p] = b
                v[1,:p] = t
                v[0,p] = h
                v[1,p] = r
    else:
        if design is None:
            print("  … a meta analysis will be performed")
            for v in it:
                sampled = ~np.isnan(v).any(axis=0)

                y=v[0,sampled]
                d=v[1,sampled]

                if (sampled).sum() > max(p+1,df):
                    b, t, h, r = meta_analysis(y=v[0], d=v[1])
                    v[0,0] = b
                    v[1,0] = t
                    v[0,1] = h
                    v[1,1] = r
                else:
                    v[...] = np.nan
        else:
            print("  … a meta regression will be performed")
            for v in it:
                sampled = ~np.isnan(v).any(axis=0)

                y=v[0,sampled]
                d=v[1,sampled]
                x=design[sampled]

                if (sampled).sum() > max(p+1,df):
                    b, t, h, r = meta_regression(y=v[0], d=v[1], x=x)
                    v[0,:p] = b
                    v[1,:p] = t
                    v[0,p] = h
                    v[1,p] = r
                else:
                    v[...] = np.nan

    result = np.zeros(mask.shape + (2,p+1,))
    result[...] = np.nan
    result[mask] = res[...,:p+1]

    return result



# Copyright 2018 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

from ..load import load

from ..protocol import filter_protocol

import pandas as pd

import numpy as np

def filter_df(df, cohort, j, paradigm):

    if cohort is None:
        cohort = slice(None)

    if paradigm is None:
        paradigm = slice(None)

    if j is None:
        j = slice(None)
    elif len(j) == 1:
        j = j[0]
    elif len(j) == 2:
        j = slice(j[0], j[1])

    df = df.loc(axis=0)[cohort, j, paradigm]

    if len(df) < 1:
        return

    df = df[ df.valid == True].copy()
    df['epi'] = df.epi.astype(int)
    df.sort_index(inplace=True)
    return df

def get_df(args, fall_back=None):
    if args.protocol:
        try:
            df = pd.read_pickle(args.protocol)
            if args.verbose:
                print('Read: {}'.format(args.protocol))
        except Exception as e:
            print('Unable to read protocol file {}'.format(args.protocol))
            print('Exception: {}'.format(e))
            return

        return filter_df(df, args.cohort, args.id, args.paradigm)

    elif args.id is not None:
        if len(args.id) > 1:
            print('When no protocol file is specified, then only one id is allowed')
            return

        try:
            date = pd.to_datetime(args.datetime, format=args.strftime)
        except Exception as e:
            print('Unable to parse date time string: {}'.format(args.datetime))
            print(e)
            return

        if hasattr(args, 'epi_code'):
            assert type(args.epi_code) is int, \
                    """

You need to provide an EPI code via --epi-code, and it must be integer,
within [-3,3], and not null.
"""
            df = pd.DataFrame({
                'cohort'   : args.cohort,
                'id'       : args.id[0],
                'date'     : date,
                'paradigm' : args.paradigm,
                'epi'      : args.epi_code,
                'valid'    : True,
                }, index=args.id)
        else:
            df = pd.DataFrame({
                'cohort'   : args.cohort,
                'id'       : args.id[0],
                'date'     : date,
                'paradigm' : args.paradigm,
                'valid'    : True,
                }, index=args.id)

        df.set_index(
                keys=['cohort', 'id', 'paradigm', 'date'],
                drop=False,
                inplace=True,
                verify_integrity=True)
        df.sort_index(inplace=True)
        df.cohort = df.cohort.astype('category')
        df.paradigm = df.paradigm.astype('category')

        return df

    elif fall_back is not None:
        try:
            instance = load(fall_back)
        except:
            try:
                instance = load(fall_back[0])
            except:
                instance = load(fall_back[1])

        if hasattr(args, 'epi_code'):
            assert type(args.epi_code) is int, \
                    """

You need to provide an EPI code via --epi-code, and it must be integer,
within [-3,3], and not null.
"""
            df = pd.DataFrame({
                'cohort'   : instance.name.cohort,
                'id'       : instance.name.j,
                'date'     : instance.name.datetime,
                'paradigm' : instance.name.paradigm,
                'epi'      : args.epi_code,
                'valid'    : True,
                }, index=[instance.name.j])
        else:
            df = pd.DataFrame({
                'cohort'   : instance.name.cohort,
                'id'       : instance.name.j,
                'date'     : instance.name.datetime,
                'paradigm' : instance.name.paradigm,
                'valid'    : True,
                }, index=[instance.name.j])

        df.set_index(
                keys=['cohort', 'id', 'paradigm', 'date'],
                drop=False,
                inplace=True,
                verify_integrity=True)
        df.sort_index(inplace=True)
        df.cohort = df.cohort.astype('category')
        df.paradigm = df.paradigm.astype('category')

        return df



# Copyright 2016-2018 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Fit a signal model to FMRI data

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

########################################################################
# Input arguments
########################################################################

    parser.add_argument('--session',
            default='../data/ses/{2}/{0}-{1:04d}-{2}-{3}.ses',
            help="""only needed if no fit is provided; """ + hp.session)

########################################################################
# Additional input arguments when a previous fit can be provided
########################################################################

    parser.add_argument('--nb-fit',
            default='../data/fit/{2}/{4}/{5}/{0}-{1:04d}-{2}-{3}-{4}-{5}.fit',
            help="""input file; if warp coefficient files are provided
            you don't need this""" + hp.sfit)

    parser.add_argument('--nb-scale-type',
            default='max',
            choices=['diagonal','max','min'],
            help="""only needed if part of the template for --nb-fit""" + hp.scale_type)

    parser.add_argument('--nb',
            default='reference',
            help="""name of the population space that was originally
            used for the fit"""  + hp.population_space)

    parser.add_argument('--ignore-fit',
            action='store_true',
            help="""ignore fit""")

########################################################################
# Output arguments
########################################################################

    parser.add_argument('--population-space',
            default='reference',
            help='output name;' + hp.population_space_name)

    parser.add_argument('--population-map',
            default='../data/pop/{2}/{4}/{0}-{1:04d}-{2}-{3}-{4}.pop',
            help='output file;' + hp.population_map)

    parser.add_argument('-o', '--protocol-log',
            default='logs/{}-fmripop.pkl',
            help=hp.protocol_log)

########################################################################
# Configuration
########################################################################

    parser.add_argument('--resolution',
            default=2.,
            type=float,
            help="""(optional) only applicable when population space is
            reference.""")

########################################################################
# Arguments specific for using the protocol API
########################################################################

    parser.add_argument('--protocol',
            help=hp.protocol)

    parser.add_argument('--cohort',
            help=hp.cohort)

    parser.add_argument('--id',
            type=int,
            nargs='+',
            help=hp.j)

    parser.add_argument('--datetime',
            help=hp.datetime)

    parser.add_argument('--paradigm',
            help=hp.paradigm)

    parser.add_argument('--strftime',
            default='%Y-%m-%d-%H%M',
            help=hp.strftime)

########################################################################
# Miscellaneous
########################################################################

    lock_handling = parser.add_mutually_exclusive_group()

    lock_handling.add_argument('-r', '--remove-lock',
            action='store_true',
            help=hp.remove_lock.format('population map'))

    lock_handling.add_argument('-i', '--ignore-lock',
            action='store_true',
            help=hp.ignore_lock.format('population map'))

    file_handling = parser.add_mutually_exclusive_group()

    file_handling.add_argument('-f', '--force',
            action='store_true',
            help=hp.force.format('population map'))

    file_handling.add_argument('-s', '--skip',
            action='store_true',
            help=hp.skip.format('population map'))

    parser.add_argument('-v', '--verbose',
            action='count',
            default=0,
            help=hp.verbose)

########################################################################
# Multiprocessing
########################################################################

    parser.add_argument('-j', '--cores',
            type=int,
            help=hp.cores)

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ..df import get_df

from ...lock import Lock

from ...load import load_result, load_session, load_population_map

from ...name import Identifier

from ...protocol import layout_dummy, layout_sdummy

from ...session import Session

from ...reference import ReferenceMaps

from ...smodel import SignalModel, Result

from ...pmap import PopulationMap, pmap_scanner, pmap_reference

import pandas as pd

import datetime

import sys

import os

from os.path import isfile, isdir, join

from multiprocessing.dummy import Pool as ThreadPool

import numpy as np

########################################################################

def call(args):
    output = args.protocol_log.format(
            datetime.datetime.now().strftime('%Y-%m-%d-%H%M'))

    if args.strftime == 'short':
        args.strftime = '%Y-%m-%d'

    ####################################################################
    # Parse protocol
    ####################################################################

    df = get_df(args, fall_back=[args.nb_fit, args.session])

    if df is None:
        sys.exit()

    ####################################################################
    # Add file layout
    ####################################################################

    df_layout = df.copy()

    layout_dummy(df_layout, 'ses',
            template=args.session,
            strftime=args.strftime
            )

    layout_sdummy(df_layout, 'fit',
            template=args.nb_fit,
            urname=args.nb,
            scale_type=args.nb_scale_type,
            strftime=args.strftime
            )

    layout_sdummy(df_layout, 'file',
            template=args.population_map,
            urname=args.population_space,
            scale_type=None,
            strftime=args.strftime
            )

    ####################################################################
    # Apply wrapper
    ####################################################################

    df['locked'] = False

    def wm(r):
        name = Identifier(cohort=r.cohort, j=r.id, datetime=r.date, paradigm=r.paradigm)

        try:
            dfile = os.path.dirname(r.file)
            if dfile and not isdir(dfile):
                os.makedirs(dfile)
        except Exception as e:
            print('{}: {}'.format(name.name(), e))

        wrapper(name              = name,
                df                = df,
                index             = r.Index,
                remove_lock       = args.remove_lock,
                ignore_lock       = args.ignore_lock,
                force             = args.force,
                skip              = args.skip,
                verbose           = args.verbose,
                file              = r.file,

                file_fit          = r.fit,
                file_ses          = r.ses,
                vb                = args.population_space,
                resolution        = args.resolution,
                ignore_fit        = args.ignore_fit,
                )

    it =  df_layout.itertuples()

    if len(df_layout) > 1 and ((args.cores is None) or (args.cores > 1)):
        try:
            pool = ThreadPool(args.cores)
            results = pool.map(wm, it)
            pool.close()
            pool.join()
        except Exception as e:
            pool.close()
            pool.terminate()
            print('Pool execution has been terminated')
            print(e)
        finally:
            files = df_layout.ix[df.locked, 'file'].values
            if len(files) > 0:
                for f in files:
                    print('Unlock: {}'.format(f))
                    os.remove(f)
            del df['locked']
    else:
        try:
            print('Process protocol entries sequentially')
            for r in it:
                wm(r)
        finally:
            files = df_layout.ix[df.locked, 'file'].values
            if len(files) > 0:
                for f in files:
                    print('Unlock: {}'.format(f))
                    os.remove(f)
            del df['locked']

    ####################################################################
    # Write protocol
    ####################################################################

    if args.verbose:
        print('Save: {}'.format(output))

    dfile = os.path.dirname(output)
    if dfile and not isdir(dfile):
       os.makedirs(dfile)

    df.to_pickle(output)

#######################################################################

def wrapper(name, df, index, remove_lock, ignore_lock, force, skip,
        verbose, file, file_fit, file_ses, vb, resolution, ignore_fit):

    if isfile(file):
        instance = load_population_map(file, name, df, index, vb, verbose)
        if type(instance) is Lock:
            if remove_lock or ignore_lock:
                if verbose:
                    print('{}: Remove lock'.format(name.name()))
                instance.unlock()
                if remove_lock:
                    return
            else:
                if verbose:
                    print('{}: Locked'.format(name.name()))
                return
        else:
            if df.ix[index,'valid'] and not force:
                if verbose:
                    print('{}: Valid'.format(name.name()))
                return
            else:
                if skip:
                    if verbose:
                        print('{}: Invalid'.format(name.name()))
                    return

    if skip:
        return

    if verbose:
        print('{}: Lock: {}'.format(name.name(), file))

    lock = Lock(name, 'fmrifit', file)
    df.ix[index, 'locked'] = True
    lock.save(file)
    df.ix[index,'valid'] = True

    ####################################################################
    # Load fit from disk
    ####################################################################

    if isfile(file_fit) and not ignore_fit:
        result = load_result(file_fit, name, df, index, vb, verbose)
        if lock.conditional_unlock(df, index, verbose):
            return

        if verbose > 1:
            print('{}: Description of the fit:'.format(result.name.name()))
            print(result.describe())

        population_map = result.population_map
        population_map.set_template(template=result.get_field('intercept', 'point'))

    else:
        session = load_session(file_ses, name, df, index, verbose)
        if lock.conditional_unlock(df, index, verbose):
            return

        if vb == 'scanner':
            population_map = pmap_scanner(session=session)
        if vb == 'reference':
            population_map = pmap_reference(session=session, resolution=resolution)

    if verbose:
        print('{}: Save: {}'.format(name.name(), file))

    try:
        population_map.save(file)
        df.ix[index,'locked'] = False
    except Exception as e:
        df.ix[index,'valid'] = False
        print('{}: Unable to write: {}'.format(name.name(), file))
        print('{}: Exception: {}'.format(name.name(), e))
        lock.conditional_unlock(df, index, verbose, True)

    if verbose > 1:
        print('{}: Description'.format(name.name()))
        print(result.describe())



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Create a block irritation instances

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

########################################################################
# Output arguments
########################################################################

    parser.add_argument('--irritation',
            default='../data/irr/{2}/{0}-{1:04d}-{2}-{3}.irr',
            help='input file;' + hp.irritation)

    parser.add_argument('-o', '--protocol-log',
            default='logs/{}-fmriirritation.pkl',
            help=hp.protocol_log)

########################################################################
# Arguments specific for the setup of an irritation instance
########################################################################

    parser.add_argument('--onsetsx',
            type=float,
            nargs='+',
            help=hp.onsetx)

    parser.add_argument('--onsetsy',
            type=float,
            nargs='+',
            help=hp.onsety)

    parser.add_argument('--durationsx',
            type=float,
            help=hp.durationx)

    parser.add_argument('--durationsy',
            type=float,
            help=hp.durationy)

    parser.add_argument('--namex',
            default='control',
            help="""name of block x""")

    parser.add_argument('--namey',
            default='stimulus',
            help="""name of block y""")

########################################################################
# If not using the protocol API, you need this
########################################################################

    parser.add_argument('--epi',
            type=int,
            help=hp.epi_code)

########################################################################
# Arguments specific for using the protocol API
########################################################################

    parser.add_argument('--protocol',
            help=hp.protocol)

    parser.add_argument('--cohort',
            help=hp.cohort)

    parser.add_argument('--id',
            type=int,
            nargs='+',
            help=hp.j)

    parser.add_argument('--datetime',
            help=hp.datetime)

    parser.add_argument('--paradigm',
            help=hp.paradigm)

    parser.add_argument('--strftime',
            default='%Y-%m-%d-%H%M',
            help=hp.strftime)

########################################################################
# Miscellaneous
########################################################################

    lock_handling = parser.add_mutually_exclusive_group()

    lock_handling.add_argument('--remove-lock',
            action='store_true',
            help=hp.remove_lock.format('population space'))

    lock_handling.add_argument('--ignore-lock',
            action='store_true',
            help=hp.ignore_lock.format('population space'))

    file_handling = parser.add_mutually_exclusive_group()

    file_handling.add_argument('-f', '--force',
            action='store_true',
            help=hp.force.format('population space'))

    file_handling.add_argument('-s', '--skip',
            action='store_true',
            help=hp.skip.format('population space'))

    parser.add_argument('-v', '--verbose',
            action='count',
            default=0,
            help=hp.verbose)

########################################################################
# Multiprocessing
########################################################################

    parser.add_argument('-j', '--cores',
            type=int,
            help=hp.cores)

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ..df import get_df

from ...lock import Lock

from ...load import load_block_irritation

from ...name import Identifier

from ...protocol import layout_dummy

from ...irritation import Block

import pandas as pd

import datetime

import sys

import os

from os.path import isfile, isdir, join

from multiprocessing.dummy import Pool as ThreadPool

import numpy as np

########################################################################

def call(args):
    output = args.protocol_log.format(
            datetime.datetime.now().strftime('%Y-%m-%d-%H%M'))

    if args.strftime == 'short':
        args.strftime = '%Y-%m-%d'

    ####################################################################
    # Parse protocol
    ####################################################################

    df = get_df(args)

    if df is None:
        sys.exit()

    ####################################################################
    # Add file layout
    ####################################################################

    df_layout = df.copy()

    layout_dummy(df_layout, 'file',
            template=args.irritation,
            strftime=args.strftime
            )

    ####################################################################
    # Apply wrapper
    ####################################################################

    df['locked'] = False

    def wm(r):
        name = Identifier(cohort=r.cohort, j=r.id, datetime=r.date, paradigm=r.paradigm)

        try:
            dfile = os.path.dirname(r.file)
            if dfile and not isdir(dfile):
                os.makedirs(dfile)
        except Exception as e:
            print('{}: {}'.format(name.name(), e))

        wrapper(name              = name,
                df                = df,
                index             = r.Index,
                remove_lock       = args.remove_lock,
                ignore_lock       = args.ignore_lock,
                force             = args.force,
                skip              = args.skip,
                verbose           = args.verbose,
                file              = r.file,

                namex    = args.namex,
                namey    = args.namey,
                onsetsx  = np.asarray(args.onsetsx),
                onsetsy  = np.asarray(args.onsetsy),
                durationsx = np.asarray(args.durationsx),
                durationsy = np.asarray(args.durationsy))

    it =  df_layout.itertuples()

    if len(df_layout) > 1 and ((args.cores is None) or (args.cores > 1)):
        try:
            pool = ThreadPool(args.cores)
            results = pool.map(wm, it)
            pool.close()
            pool.join()
        except Exception as e:
            pool.close()
            pool.terminate()
            print('Pool execution has been terminated')
            print(e)
        finally:
            files = df_layout.ix[df.locked, 'file'].values
            if len(files) > 0:
                for f in files:
                    print('Unlock: {}'.format(f))
                    os.remove(f)
            del df['locked']
    else:
        try:
            print('Process protocol entries sequentially')
            for r in it:
                wm(r)
        finally:
            files = df_layout.ix[df.locked, 'file'].values
            if len(files) > 0:
                for f in files:
                    print('Unlock: {}'.format(f))
                    os.remove(f)
            del df['locked']

    ####################################################################
    # Write protocol
    ####################################################################

    if args.verbose:
        print('Save: {}'.format(output))

    dfile = os.path.dirname(output)
    if dfile and not isdir(dfile):
       os.makedirs(dfile)

    df.to_pickle(output)

    #if args.epi_code is None:
    #    print('Warning: protocol has not been equipped with a valid EPI code')

########################################################################

def wrapper(name, df, index, remove_lock, ignore_lock, force, skip,
        verbose, file, namex, namey, onsetsx, onsetsy, durationsx,
        durationsy):

    if isfile(file):
        instance = load_block_irritation(file, name, df, index, verbose)
        if type(instance) is Lock:
            if remove_lock or ignore_lock:
                if verbose:
                    print('{}: Remove lock'.format(name.name()))
                instance.unlock()
                if remove_lock:
                    return
            else:
                if verbose:
                    print('{}: Locked'.format(name.name()))
                return
        else:
            if df.ix[index,'valid'] and not force:
                if verbose:
                    print('{}: Valid'.format(name.name()))
                return
            else:
                if skip:
                    if verbose:
                        print('{}: Invalid'.format(name.name()))
                    return

    if skip:
        return

    if verbose:
        print('{}: Lock: {}'.format(name.name(), file))

    lock = Lock(name, 'fmrifit', file)
    df.ix[index, 'locked'] = True
    lock.save(file)
    df.ix[index,'valid'] = True

    ####################################################################
    # Create irritation instance
    ####################################################################

    try:
        irritation = Block(name=name,
                names=[namex, namey],
                onsets={namex:onsetsx,namey:onsetsy},
                durations={namex:durationsx,namey:durationsy})

        if verbose:
            print('{}: Save: {}'.format(name.name(), file))

        irritation.save(file)
        df.ix[index,'locked'] = False

    except Exception as e:
        df.ix[index,'valid'] = False
        print('{}: Unable to create: {}'.format(name.name(), file))
        print('{}: Exception: {}'.format(name.name(), e))
        lock.conditional_unlock(df, index, verbose, True)
        return



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Defines a class for the FMRI signal model and its fits.

"""

from .affines import Affine

from .name import Identifier

from .session import Session

from .reference import ReferenceMaps

from .diffeomorphisms import Image

from .pmap import PopulationMap

from .irritation import Irritation

from .fit import fit_field, extract_field, model_at, data_at

import time

import numpy as np

from numpy.linalg import inv

from numpy.linalg import norm

import scipy.stats.distributions as dist

from patsy import dmatrix

from pandas import DataFrame

import pickle

import math

class SignalModel:
    """
    The signal model

    Parameters
    ----------
    session : Session
        A session (with fitted foreground).
    reference_maps : ReferenceMaps
        The estimated reference maps.
    population_map : PopulationMap
        A population map
    """
    def __init__(self, session, reference_maps, population_map):
        assert type(session) is Session, 'session must be of type Session'
        assert type(reference_maps) is ReferenceMaps, 'reference_maps must be of type ReferenceMaps'
        assert type(population_map) is PopulationMap, 'population_map must be of type PopulationMap'

        self.session = session
        self.irritation = session.irritation
        self.reference_maps = reference_maps
        self.population_map = population_map

        self.shape = (session.numob, session.shape[session.ep])
        self.slice_timing_design = session.slice_time

        self.name = self.session.name
        self.epi_code = self.session.epi_code
        self.ep = abs(self.epi_code) - 1

        # Affine maps from the index space of the Session to
        # the reference space of the ReferenceMaps.

        # TODO: when having added the option: --reference-maps None,
        # replace this with something reasonable
        self.references = reference_maps.scan_references.inv().dot(session.reference)

        self.tsr = self.session.temporal_resolution / self.shape[1]

    #######################################################################
    # Set hyperparameters for the fit
    #######################################################################

    def set_hyperparameters(self, scale_type='max', factor=3,
            scale=None, mass=None):
        """
        Set up hyperparameters for field estimation

        All parameters which control details the effect field estimation
        and which are not per se sample design specific are called
        Hyperparameters.

        Parameters
        ----------
        scale_type : str
            One of min, max (default), diagonal
        scale : float
            Standard deviation of the Gaussian used to weight
            observations with respect to distance to point.
        mass : float (optional)

        Notes
        -----
        The only hyperparameter you need to set is `scale`, which is set
        for you if you set the scale_type.

        Factor takes precedence to mass, as I will assume if you are
        setting the mass, you know what you are doing.
        """
        if scale is None:
            if scale_type == 'diagonal':
                self.scale_type = scale_type
                self.scale = 0.5 * self.session.reference.diagonal()
            elif scale_type == 'min':
                self.scale_type = scale_type
                self.scale = 0.5 * min(self.session.reference.resolution())
            elif scale_type == 'max':
                self.scale_type = scale_type
                self.scale = 0.5 * max(self.session.reference.resolution())
            else:
                raise ValueError('If SCALE is not set, SCALE_TYPE is compulsory.')
        else:
            self.scale_type = 'user'
            self.scale = scale

        if factor is not None:
            assert factor > 0, 'factor must be strictly positve'
            self.factor = factor
            self.radius = factor * self.scale
            self.mass   = dist.norm.cdf(factor) - dist.norm.cdf(-factor)

        if mass is not None:
            assert mass > 0, 'mass must be strictly positive'
            assert mass < 1, 'mass must be strictly less than 1'
            self.mass = mass
            self.factor = dist.norm.ppf(1 - (1-mass)/2)
            self.radius = self.factor * self.scale

    #######################################################################
    # The mean index affine
    #######################################################################

    def mean_index_affine(self):
        """
        Affine transformation from reference space to the index space of
        the acquisition grid

        Returns
        -------
        The affine transformation that maps a coordinate in reference
        space to the mean index in the index space of the acquisition
        grid of the session.
        """
        return self.references.inv().mean()

    #######################################################################
    # All about the observations
    #######################################################################

    def coordinates(self):
        """
        The coordinate grid of observations

        The coordinates in subject reference space of all points at
        which MR signals have been aquired.

        Returns
        -------
        ndarray, shape (n,x,y,z,3), dtype: float
        """
        n,x,y,z = self.session.data.shape
        indices = ((slice(0,x), slice(0,y), slice(0,z)))

        coordinates = np.empty((n,x,y,z,3))
        for t in range(n):
            coordinates[t] = self.references.index(t).apply_to_indices(indices)

        return coordinates

    def create_slice_timing_design(self):
        return self.slice_timing_design

    def create_irritation_design(self, **kwargs):
        assert hasattr(self, 'slice_timing_design'), 'first create the slice timing design'
        self.irritation_design = self.irritation.design(
                slice_timing=self.slice_timing_design, **kwargs)
        return self.irritation_design

    def observations(self):
        """
        Create observation matrix

        Returns
        -------
        ndarray, shape (n,x,y,z,6), dtype: float
            [...,:3] = coordinates of observation
            [..., 3] = MR signal response
            [..., 4] = time of observation
            [..., 5] = task during time of observation
            [..., 6] = block number during time of observation
            [..., 7] = scan cycle
            [..., 8] = slice number
        """
        assert hasattr(self, 'slice_timing_design'), 'first set the slice time design'
        assert hasattr(self, 'irritation_design'), 'first set the irritation design'
        n,x,y,z = self.session.data.shape

        observations = np.zeros((n,x,y,z,9))
        observations[...,:3] = self.coordinates()
        observations[..., 3] = self.session.data

        tmp = np.moveaxis(observations, self.ep+1, 1)
        tmp[...,4]  = self.slice_timing_design[...,None,None]
        tmp[...,5:7] = self.irritation_design[...,None,None,:]

        x = np.mgrid[:self.shape[0], :self.shape[1]]
        x = np.moveaxis(x, 0, -1)
        tmp[...,7:9] = x[...,None,None,:]

        return observations

    def create_data_matrix(self, burn_in=None, verbose=True):
        """
        Create the data matrix

        This is the essentially the same matrix as returned by
        .observations(). However, scan cycles with no subject
        irritation or missing data are removed. However, time is centred
        to its mean.

        Returns
        -------
        ndarray, shape (n',x,y,z,6), dtype: float
            [...,:3] = coordinates of observation
            [..., 3] = MR signal response
            [..., 4] = time of observation
            [..., 5] = task during time of observation
            [..., 6] = block number during time of observation
        """
        self.burn_in = burn_in

        observations = self.observations()

        # remove outlying scans due to severe movements of the subject
        # in the scanner

        if hasattr(self.reference_maps, 'outlying_scans'):
            scans = self.reference_maps.outlying_scans
            if scans.any():
                tmp = np.moveaxis(observations, self.ep+1, 1)
                tmp [scans] = np.nan

                if verbose:
                    print('{}: Removed {} ({:.2f}%) outlying scans'.format(
                        self.name.name(), scans.sum(), 100*scans.mean()))

        # these observations have either no response (e.g. they are
        # outside of the brain) or there are not all covariates
        # available (e.g. they have been acquired outside of task blocks
        # or the position of the slice in space could not be detected
        # with success.
        none = np.isnan(observations).any(axis=-1)

        # these observations have no response as they are identical
        # zero.
        null = np.isclose(observations[...,3], 0)

        # put them together and set all observations to NAN when
        # anything is missing or invalid
        missings = none | null
        observations[missings] = np.nan

        # scan cycles that we do not need to process at all
        if burn_in:
            observations[:burn_in] = np.nan

        # it is more numerically stable to work with a demeaned time
        # vector. This has also the consequence that the intercept will
        # refer to the mean signal intensity at the midpoint of the FMRI
        # session. Again, this is the point at which the intercept will
        # have the least variance.
        self.midpoint = np.nanmean(observations[...,4])
        observations[...,4] = observations[...,4] - self.midpoint

        if verbose:
            valid = np.isfinite(observations).all(axis=-1)
            print("""{}:
            …voxels which are part of the model: {:>10,d}
            …voxels which are not:               {:>10,d}
            …intercept field refers to:          {:>10.2f} s
            """.format(self.name.name(), valid.sum(), (~valid).sum(), self.midpoint))

        data_matrix = DataFrame({
            'i'     : observations[...,0].ravel(),
            'j'     : observations[...,1].ravel(),
            'k'     : observations[...,2].ravel(),
            'y'     : observations[...,3].ravel(),
            'time'  : observations[...,4].ravel(),
            'task'  : observations[...,5].ravel(),
            'block' : observations[...,6].ravel(),
            'cycle' : observations[...,7].ravel(),
            'slice' : observations[...,8].ravel()})

        data_matrix.dropna(inplace=True)
        self.data_matrix = data_matrix
        return self.data_matrix

    def create_design_matrix(self, formula='C(task)/C(block, Sum)',
            parameter=['intercept', 'task'], verbose=True):
        """
        Create the design matrix

        Returns
        -------
        (DesignMatrix, DesignMatrix)
        """
        assert hasattr(self, 'data_matrix'), 'first set the design using .create_data_matrix()'

        dmat = dmatrix(formula, self.data_matrix, eval_env=-1)
        names = dmat.design_info.column_names
        parameter_dict = { p : [p in n.lower() for n in names].index(True) for p in parameter}

        self.exog  = np.asarray(dmat)
        self.formula = formula
        self.parameter_dict = parameter_dict
        return dmat

    ####################################################################
    # Fit at one coordinate
    ####################################################################

    def data_at_subject_coordinate(self, x):
        """
        Fit the signal model to data at specificed coordinates given
        with respect to the coordinate system of the subject reference
        space.

        Parameters
        ----------
        x : ndarray, shape (3,), dtype: float
            The coordinates at which to fit the model
        """
        assert hasattr(self, 'scale'), 'first set hyperparameters'
        assert hasattr(self, 'radius'), 'first set hyperparameters'
        assert hasattr(self, 'data_matrix'), 'first set data_matrix'

        return data_at(coordinate=x,
                data=self.data_matrix,
                scale=self.scale,
                radius=self.radius)

    def data_at_index(self, index):
        """
        Fit the signal model to data at specificed coordinates given
        with respect to the coordinate system of the subject reference
        space.

        Parameters
        ----------
        x : ndarray, shape (3,), dtype: float
            The coordinates at which to fit the model
        """
        x=self.population_map.diffeomorphism.apply_to_index(index)
        return self.data_at_subject_coordinate(x)

    def model_at_subject_coordinate(self, x, formula='y~C(task)/C(block, Sum)'):
        """
        Fit the signal model to data at specificed coordinates given
        with respect to the coordinate system of the subject reference
        space.

        Parameters
        ----------
        x : ndarray, shape (3,), dtype: float
            The coordinates at which to fit the model
        """
        assert hasattr(self, 'scale'), 'first set hyperparameters'
        assert hasattr(self, 'radius'), 'first set hyperparameters'
        assert hasattr(self, 'data_matrix'), 'first set data_matrix'

        return model_at(coordinate=x, formula=formula,
                data=self.data_matrix,
                scale=self.scale,
                radius=self.radius)

    def model_at_index(self, index, **kwargs):
        """
        Fit the signal model to data at specificed coordinates

        Parameters
        ----------
        index : tuple
            The index at which to fit the model
        """
        x=self.population_map.diffeomorphism.apply_to_index(index)
        return self.model_at_subject_coordinate(x, **kwargs)

    def model_at_coordinate(self, x, **kwargs):
        """
        Fit the signal model to data at specificed coordinates given
        with respect to the coordinate system of the population space.

        Parameters
        ----------
        x : ndarray, shape (3,), dtype: float
            The coordinates at which to fit the model
        """
        x=self.population_map.diffeomorphism.apply(x)
        return self.model_at_subject_coordinate(x, **kwargs)

    ####################################################################
    # Fit at many coordinates
    ####################################################################

    def get_data_mask(self, verbose=True):
        coordinates = self.population_map.diffeomorphism.coordinates()
        to_index = self.mean_index_affine()
        idx = to_index.apply(coordinates)
        idx = idx.round().astype(int)

        mdx = idx.reshape((-1,3))

        di, dj, dk = self.session.shape

        for x in iter(mdx):
            i,j,k = x
            if i < 0 or j < 0 or k < 0 or i >= di or j >= dj or k >= dk:
                x[...] = 0
            elif np.isfinite(self.session.data[:,i,j,k]).any():
                x[...] = 1
            else:
                x[...] = 0

        mask = (idx > 0).any(-1)

        if verbose:
            print("""{}:
            …coordinates which are in data mask: {:>10,d}
            …coordinates which are not:          {:>10,d}""".format(
                self.name.name(), mask.sum(), (~mask).sum()))

        return mask

    def create_estimation_matrix(self, mask=True, verbose=True):
        """
        Parameters
        ----------
        mask : None or bool or str or ndarray, dtype: bool
            string can be one of 'template', 'template_mask', or 'target'
        verbose : bool
            increase output verbosity

        Returns
        -------
        ndarray
        """
        coordinates = self.population_map.diffeomorphism.coordinates()

        if (mask is False) or (mask is None):
            mask = None

        elif mask is True:
            if hasattr(self.population_map, 'template_mask'):
                mask = self.population_map.template_mask.get_mask()
                maskname = 'template mask'
            elif hasattr(self.population_map, 'template'):
                mask = self.population_map.template.get_mask()
                maskname = 'template'
            else:
                mask = None
                maskname = 'data mask'

        elif type(mask) is str:
            if mask == 'template_mask':
                mask = self.population_map.template_mask.get_mask()
                massname = 'template mask'
            elif mask == 'template':
                mask = self.population_map.template.get_mask()
                maskname = 'template'
            elif mask == 'target':
                mask = self.population_map.target.get_mask()
                masskname = 'target'
            else:
                mask = None
                maskname = 'data mask'

        else:
            maskname = 'user defined'

        if mask is not None:
            assert type(mask) is np.ndarray, 'mask must be an ndarray'
            assert mask.dtype == bool, 'mask must be of dtype bool'
            assert mask.shape == coordinates.shape[:-1], 'mask shape must match image shape'
            mask = mask & self.get_data_mask()
        else:
            mask = self.get_data_mask()

        if verbose:
            print('{}: Fit is restricted to: {}'.format(self.name.name(), maskname))

        self.estimation_matrix = coordinates
        self.estimation_mask   = mask
        return self.estimation_matrix, self.estimation_mask

    def fit_at_subject_coordinates(self, coordinates, mask=None, verbose=True):
        """
        Fit the signal model to data

        Parameters
        ----------
        coordinates : None or ndarray, shape (…,3), dtype: float
            The coordinates at which to fit the model
        verbose : bool
            increase output verbosity

        Returns
        -------
        Result
        """
        assert hasattr(self, 'scale'), 'first set hyperparameters using .set_hyperparameters()'
        assert hasattr(self, 'radius'), 'first set hyperparameters using .set_hyperparameters()'
        assert hasattr(self, 'data_matrix'), 'first set the design using .create_data_matrix()'
        assert hasattr(self, 'exog'), 'first set the design using .create_design_matrix()'

        if verbose:
            print('{}: Start fit'.format(self.name.name()))

        old_settings = np.seterr(divide='raise', invalid='raise')
        time0 = time.time()

        statistics, parameter_dict, value_dict = fit_field(
                coordinates = coordinates,
                mask        = mask,
                endog       = self.data_matrix.y.values,
                exog        = self.exog,
                agc         = self.data_matrix[['i','j','k']].values,
                scale       = self.scale,
                radius      = self.radius,
                verbose     = verbose,
                name        = self.name.name())

        time1 = time.time()
        np.seterr(**old_settings)

        if verbose:
            time_spend = time1 - time0
            print('{}: Time needed for the fit: {:.2f} min'.format(
                self.name.name(), time_spend / 60))
            print('{}: Time needed for the fit: {:.2f} h'  .format(
                self.name.name(), time_spend / 60**2))

        parameter_dict.update(self.parameter_dict)

        result = Result(
                coordinates     = coordinates,
                statistics      = statistics,
                population_map  = self.population_map,
                hyperparameters = self.hyperparameters(),
                parameter_dict  = parameter_dict,
                value_dict      = value_dict)

        return result

    def fit_at_indices(self, indices, **kwargs):
        coordinates = self.diffeomorphism.apply_to_indices(indices)
        return self.fit_at_subject_coordinates(coordinates = coordinates, **kwargs)

    def fit(self, mask=True, verbose=True):
        """
        Fit the signal model to data

        Parameters
        ----------
        mask : None or bool or str or ndarray, dtype: bool
            string can be one of 'template', 'template_mask', or 'target'
        verbose : bool
            increase output verbosity

        Returns
        -------
        Result : Fitted field.
        """
        if not hasattr(self, 'scale'):
            if verbose:
                print('{}! Set hyperparameters to default'.format(self.name.name()))
            self.set_hyperparameters()

        if not hasattr(self, 'radius'):
            if verbose:
                print('{}! Set hyperparameters to default'.format(self.name.name()))
            self.set_hyperparameters()

        if not hasattr(self, 'data_matrix'):
            if verbose:
                print('{}! Set data matrix to default (no burn in)'.format(self.name.name()))
            self.create_data_matrix()

        if not hasattr(self, 'exog'):
            if verbose:
                print('{}! Set design matrix / model to default'.format(self.name.name()))
            self.create_design_matrix()

        if verbose:
            print('{}: Create estimation matrix and mask'.format(self.name.name()))
        self.create_estimation_matrix(mask=mask, verbose=verbose)

        return self.fit_at_subject_coordinates(
                coordinates = self.estimation_matrix,
                mask        = self.estimation_mask)


    ###################################################################
    # Descriptive statistics of this session
    ###################################################################

    def hyperparameters(self):
        return {
                'scale_type':self.scale_type,
                'scale':self.scale,
                'factor':self.factor,
                'mass':self.mass,
                'radius':self.radius,
                #'half_edge':self.half_edge,
                }

    def describe(self):
        description = """
        Hyperparameters were set to:
            Scale type:   {:s}
            Scale:        {:.2f} mm
            Factor:       {:.2f}
        Resulting in:
            Mass:         {:3>,.5f}
            FWHM:         {:3>,.2f} mm
            Radius:       {:3>,.2f} mm
            Diagonal:     {:3>,.2f} mm"""
        hyperparameters = self.hyperparameters()
        return description.format(
                hyperparameters['scale_type'],
                hyperparameters['scale'],
                hyperparameters['factor'],

                hyperparameters['mass'],
                2*np.sqrt(2*np.log(2)) * hyperparameters['scale'],
                hyperparameters['radius'],
                2*hyperparameters['radius'],
                )

    def save(self, file, **kwargs):
        """
        Save instance to disk

        This will save the current instance to disk for later use.

        Parameters
        ----------
        file : str
            File name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)

#######################################################################
#######################################################################
#
# Store the result of a fit to the data of an FMRI experiment
#
#######################################################################
#######################################################################

class Result:
    """
    Result of a FMRI fitting

    Defines a class for the result, i.e., the estimated effect field of
    an FMRI experiments, fitted by the function attribute of the `FMRI`
    class.

    Parameters
    ----------
    name : Identifier
        An identifier
    population_map : PopulationMap
    field : ndarray, shape (…,3), dtype: float
        The effect field, i.e., result of fitting the RSM model.
    coordinates : ndarray, shape (…,3), dtype: float
    hyperparameters : dict
    parameters : dict {str: int}
        The names of the parameters which have been fitted by the
        signal model in the order of appearance in the model.
    """
    def __init__(self, coordinates, statistics, population_map,
            hyperparameters, parameter_dict, value_dict):
        assert type(coordinates) is np.ndarray, 'coordinates must be a ndarray'
        assert type(population_map) is PopulationMap, 'population_map must be a PopulationMap'
        assert type(hyperparameters) is dict, 'hyperparameters must be a dict'
        assert type(parameter_dict) is dict, 'parameter_dict must be a dict'
        assert type(value_dict) is dict, 'value_dict must be a dict'

        self.coordinates = coordinates
        self.statistics = statistics
        self.population_map = population_map
        self.hyperparameters = hyperparameters
        self.parameter_dict = parameter_dict
        self.value_dict = value_dict

        self.name = self.population_map.name

    ####################################################################
    # Extract summary statistics
    ####################################################################

    def get_field(self, param, value='other'):
        """
        Extract scalar field

        Parameters
        ----------
        param : str
            Check the attribute parameter_dict for possible values of
            `param`.
        value : str
            Check the attribute value_dict for possible values of
            `value`. Note that 'tstatistic' is also valid and will be
            calculated from the key point and stderr.

        Returns
        -------
        Image
            The queried field

        Notes
        -----
        It only makes sense to combine intercept or activation
        with point, stderr, and tstatistic, and it only makes sense to
        combine other with mse (an estimate of sigma-squared) or df
        (residual degrees of freedom)
        """
        #if value == 'tstatistic':
        #    pfield = extract_field(
        #            field=self.statistics,
        #            param=param,
        #            value='point',
        #            parameter_dict=self.parameter_dict,
        #            value_dict=self.value_dict)

        #    sfield = extract_field(
        #            field=self.statistics,
        #            param=param,
        #            value='stderr',
        #            parameter_dict=self.parameter_dict,
        #            value_dict=self.value_dict)

        #    field = pfield / sfield

        #else:
        #    field = extract_field(
        #            field=self.statistics,
        #            param=param,
        #            value=value,
        #            parameter_dict=self.parameter_dict,
        #            value_dict=self.value_dict)

        field = extract_field(
                field=self.statistics,
                param=param,
                value=value,
                parameter_dict=self.parameter_dict,
                value_dict=self.value_dict)

        return Image(
                reference=self.population_map.diffeomorphism.reference,
                data=field,
                name='-'.join((param,value)))

    def volume(self):
        return np.isfinite(self.statistics).sum() * \
                self.population_map.diffeomorphism.reference.volume()

    ###################################################################
    # Descriptive statistics of this session
    ###################################################################

    def mask(self, mask=True, verbose=False):
        """
        Apply mask to parameter fields
        """
        if (mask is False) or (mask is None):
            mask = None

        elif mask is True:
            if hasattr(self.population_map, 'template_mask'):
                mask = self.population_map.template_mask.get_mask()
                maskname = 'template mask'
            elif hasattr(self.population_map, 'template'):
                mask = self.population_map.template.get_mask()
                maskname = 'template'
            else:
                mask = None
                maskname = 'none'

        elif type(mask) is str:
            if mask == 'template_mask':
                mask = self.population_map.template_mask.get_mask()
                massname = 'template mask'
            elif mask == 'template':
                mask = self.population_map.template.get_mask()
                maskname = 'template'
            elif mask == 'target':
                mask = self.population_map.target.get_mask()
                masskname = 'target'
            else:
                mask = None
                maskname = 'none'

        else:
            maskname = 'user defined'

        if verbose:
            print('{}: Fit is restricted to: {}'.format(self.name.name(), maskname))

        if mask is not None:
            assert type(mask) is np.ndarray, 'mask must be an ndarray'
            assert mask.dtype == bool, 'mask must be of dtype bool'
            assert mask.shape == self.statistics.shape[:-2], 'mask shape must match image shape'
            self.statistics [ ~mask ] = np.nan

    def summary(self, quantile_t=None):
        tf = self.get_field('activation', 'tstatistic')

        if quantile_t:
            assert quantile_t <= 1, 'quantile_t must be less than 1'
            assert quantile_t >= 0, 'quantile_t must be greater than 0'
            threshold_t = np.nanpercentile(tf.data, q=quantile_t*100)
            neuronal_active = tf.data > threshold_t
        else:
            neuronal_active = np.isfinite(tf.data)

        intercept  = self.parameter_dict ['intercept']
        activation = self.parameter_dict ['activation']
        degfreedom = self.parameter_dict ['df']

        point      = self.value_dict ['point']
        stderr     = self.value_dict ['stderr']
        tstatistic = self.value_dict ['tstatistic']
        other      = self.value_dict ['other']

        statistics = self.statistics [neuronal_active]

        return DataFrame({
            'cohort'   : self.name.cohort,
            'id'       : self.name.j,
            'paradigm' : self.name.paradigm,
            'date'     : self.name.datetime,
            'x' : self.coordinates[neuronal_active,0],
            'y' : self.coordinates[neuronal_active,1],
            'z' : self.coordinates[neuronal_active,2],
            'intercept'  : statistics[...,point,intercept],
            'activation' : statistics[...,point,activation],
            'stderr'     : statistics[...,stderr,activation],
            'tstatistic' : statistics[...,tstatistic,activation],
            'df' : statistics[...,other,degfreedom],
            })

    def descriptive_statistics(self):
        mask = np.isnan(self.statistics).any(axis=(-1,-2))
        nooc = np.isnan(mask).all(axis=(0,1))
        return self.volume(), (~nooc).sum()

    def describe(self):
        description = """
        Cohort:   {}
        Subject:  {}
        Paradigm: {}
        Hyperparameters were set to:
            Scale type:      {:s}
            Factor:          {:.2f}
        Resulting in:
            Scale:           {:.2f} mm
            Mass:            {:.5f}
            FWHM:            {:.2f} mm
            Radius:          {:.2f} mm
            Diagonal:        {:.2f} mm
        Shape:               {}
        Volume:              {:.2f} mm^3
        """
        return description.format(
                self.population_map.diffeomorphism.nb.cohort,
                self.population_map.diffeomorphism.nb.j,
                self.population_map.diffeomorphism.nb.paradigm,

                self.hyperparameters['scale_type'],
                self.hyperparameters['factor'],

                self.hyperparameters['scale'],
                self.hyperparameters['mass'],
                2*np.sqrt(2*np.log(2)) * self.hyperparameters['scale'],
                self.hyperparameters['radius'],
                2*self.hyperparameters['radius'],

                self.statistics.shape,
                self.volume())

    #######################################################################
    # Save instance to disk
    #######################################################################

    def save(self, file, **kwargs):
        """
        Save instance to disk

        This will save the current instance to disk for later use.

        Parameters
        ----------
        file : str
            File name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Sample

"""

from .load import load

from .diffeomorphisms import Image

import numpy as np

import pickle

from pandas import DataFrame

class Sample:
    """
    Sampled activation fields of a FMRI study
    """
    def __init__(self, population_space, covariates, statistics):
        """
        Parameters
        ----------
        population_space : Image
        covariates : DataFrame
        statistics : ndarray, shape (…,3)
        """
        assert type(population_space) is Image, 'population_space must be of type Image'
        self.population_space = population_space

        assert type(covariates) is DataFrame
        covariates.reset_index(inplace=True, drop=True)
        self.covariates = covariates

        assert statistics.shape == population_space.shape + (2,len(covariates))
        self.statistics = statistics

    def filter(self, b=None):
        """
        Notes
        -----
        Here, b should be a slice object, you cannot work with the index
        of the covariate data frame, but must use integer location
        indices instead.
        """
        if b is None:
            b = self.covariates.valid

        if b.dtype == np.dtype(bool):
            covariates = self.covariates.ix[b]
        else:
            covariates = self.covariates.iloc[b]

        return Sample(
                population_space = self.population_space,
                covariates       = covariates.ix[b],
                statistics       = self.statistics[...,b])

    ###################################################################
    # Description
    ###################################################################

    def describe(self):
        description = """
No. of samples: {:d}
{}
        """
        valid = self.covariates.groupby(
                ['cohort','paradigm','valid']).id.agg(['count'])
        return description.format(
                len(self.covariates), valid)

    ####################################################################
    # Save nstance to and from disk
    ####################################################################

    def save(self, file, **kwargs):
        """
        Save model instance to disk

        This will save the current model instance to disk for later use.

        Parameters
        ----------
        file : str
            File name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)



# Copyright 2016-2018 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Fit signal model to signal data

"""

import statsmodels.api as sm

import statsmodels.formula.api as smf

import numpy as np

########################################################################

def extract_field(field, param, value, parameter_dict, value_dict):
    return field[..., value_dict[value], parameter_dict[param]]

def fit_field(coordinates, mask, endog, exog, agc, scale:float,
        radius:float, verbose=True, name=None):

    assert agc.shape[0] == endog.shape[0], 'shapes do not match'
    assert agc.shape[0] == exog.shape[0], 'shapes do not match'

    double_squared_scale = -2*scale**2

    p = exog.shape[1]
    result = np.zeros(coordinates.shape[:-1] + (4,max(p,3),), dtype=float)
    result[...,0,:3] = coordinates.copy()

    if mask is None:
        points_to_estimate = result.reshape((-1,) + result.shape[-2:])
        partial_fit = False

        if verbose:
            print("""{}:
            …coordinates in the domain to estimate: {:>10,d}""".format(
                name, points_to_estimate.shape[0]))
    else:
        assert type(mask) is np.ndarray, 'mask must be an ndarray'
        assert mask.dtype == bool, \
                'mask is not None it must be of dtype bool'
        assert mask.shape == coordinates.shape[:-1], \
                'image dimensions of coordinates and mask must match'
        result[~mask] = np.nan
        points_to_estimate = result[mask]
        partial_fit = True

        if verbose:
            print("""{}:
            …coordinates in the domain to estimate: {:>10,d}
            …coordinates in the domain not to:      {:>10,d}""".format(
                name, points_to_estimate.shape[0], (~mask).sum()))

    for x in iter(points_to_estimate):
        distances = ((agc - x[...,0,:3])**2).sum(axis=1)
        valid = np.less(distances, radius**2)
        if valid.sum() > 120:
            try:
                fit = sm.WLS(
                    endog    = endog[valid],
                    exog     = exog[valid],
                    weights  = np.exp(distances[valid] / double_squared_scale),
                    hasconst = True).fit()
                x[0]   = fit.params
                x[1]   = fit.bse
                x[2]   = fit.tvalues
                x[3,0] = fit.mse_resid
                x[3,1] = fit.df_resid
            except Exception as e:
                print("""{}: Exception at {}: {}""".format(
                    name, x[...,0,:3], e))
                x[...] = np.nan
        else:
            x[...] = np.nan

    if partial_fit:
        result[mask] = points_to_estimate

    value_dict = {'point':0, 'stderr':1, 'tstatistic':2, 'other':3}
    parameter_dict = {'mse':0, 'df':1}

    return result, parameter_dict, value_dict

def data_at(coordinate, data, scale : float, radius : float):
    agc = data[['i','j','k']].values
    distances = ((agc - coordinate)**2).sum(axis=1)
    valid = np.less(distances, radius**2)

    double_squared_scale = -2*(scale**2)
    weights  = np.exp( distances[valid] / double_squared_scale )

    data = data[valid].copy()
    data['weight'] = weights
    return data

def model_at(formula, **kwargs):
    data = data_at(**kwargs)
    return smf.wls(formula, weights=data.weight, data=data)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Code for tracking a rigid body which has been sampled at a fixed lattice
in $ℝ^3$ by the principle axis method.

"""

import numpy as np

from fmristats.affines import Affine, Affines, cartesian2homogeneous

def fit_by_pca(data, reference):
    """
    Tracking a rigid body

    This will track a rigid body, which has been sampled at a fixed
    lattice in :math:`ℝ^3`, by the so-called principle axis method.

    Parameters
    ----------
    data : ndarray, shape (n,x,y,z)
        Matrix of observations
    reference : Affine
        Affine transformation that maps an index to (physical) position

    Returns
    -------
    scan_references : Affines
        Rigid body transformations mapping from the given coordinate
        system (specified by `reference`_) **to** the coordinate system
        of the body, in which zero corresponds to the bary centre of the
        body and the base axis correspond to its principle axis'.

        This is the other way around!
    w : ndarray, shape (n,3)
        Principle semi axis lengths of the rigid body.

    Notes
    -----
    The function implements a (or the) principle axis method for rigid
    body tracking.  The method fits a orthonormal base (i.e. a
    coordinate system) onto data, such that the base vectors correspond
    to the principle axis (v) of the body.  The returned scan_references
    are affine transformations that map from the specified coordinate
    system to the coordinates of these points in the body.
    """
    numob, i, j, k = data.shape
    data = data.copy()
    data [ np.isnan(data) ] = 0

    n,x,y,z = data.shape
    indices = ((slice(0,x), slice(0,y), slice(0,z)))
    lattice = reference.apply_to_indices(indices)
    lattice = np.moveaxis(lattice, -1 ,0)

    # centre of masses
    com = np.empty((3,numob))
    for t in range(numob):
        com[:,t] = (data[t] * lattice).sum(axis=(1,2,3)) / data[t].sum()

    # demeaned lattice
    dm = np.repeat(lattice[...,None], numob, axis=-1) - com[:,None,None,None,:]

    # unweighted variance-covariance for each time point
    cov = dm[:,None] * dm

    # variance-covariance
    var = np.empty((3,3,numob))
    for t in range(numob):
        var[:,:,t] = (data[t] * cov[...,t]).sum(axis=(2,3,4)) / data[t].sum()

    # eigen decomposition: eigenvalues in w, eigenvectors in v
    var = np.moveaxis(var, -1, 0)
    w, v = np.linalg.eig(var)

    # scan references
    scan_references = np.empty((numob,4,4))
    for t in range(numob):
        scan_references[t] = cartesian2homogeneous(v[t], com[:,t])

    return Affines(scan_references), w



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

from . import Image, Session

import numpy as np

import nibabel as ni

#######################################################################
# Image
#######################################################################

def nii2image(nii, name=None):
    """
    Converts Nifti1Image to Image
    """
    data = nii.get_data()
    assert len(data.shape) == 3, 'not a 3D-image'
    return Image(reference=nii.affine, data=data, name=name)

def image2nii(image):
    """
    Converts Image to Nifti1Image
    """
    return ni.Nifti1Image(image.data, image.reference.affine)

#######################################################################
# Session
#######################################################################

def nii2session(name, nii, epi_code):
    """
    Create a session instance from a Nifti1Image

    Parameters
    ----------
    name : str
    nii : nibabel.nifti1.Nifti1Image
    plain : int
    """
    zooms = nii.header.get_zooms()
    return Session(name=name,
            data  = np.rollaxis(nii.get_data(), -1),
            epi_code = epi_code,
            spacial_resolution = zooms[:-1],
            temporal_resolution = zooms[-1],
            reference = nii.affine)



import pandas as pd

from pandas import Series, DataFrame

# TODO: give also the option to provide a list of ids!

def filter_protocol(df, cohort=None, j=None, k=None, paradigm=None):
    """
    Filter protocol data frame to specified cohort, within cohort id,
    and paradigm.

    Parameters
    ----------
    cohort : str
        Cohort name
    j : int
        Within cohort id of a subject
    paradigm : str
        Paradigm name of the stimulus

    Notes
    -----
    Return a copy of the specified data frame that only contains rows
    which fit to the given cohort name, within cohort id, and paradigm.
    Additionally only rows which marked as valid remain in the data
    frame.

    Returns
    -------
    The filtert data frame
    """
    if cohort:
        assert cohort in df.cohort.cat.categories, '{} is not a listed cohort'.format(cohort)
        df = df[(df.cohort == cohort)]

    if k:
        assert not (cohort is None), 'if k is specified, cohort must be specified'
        assert not (j is None), 'if k is specified, j must be specified'
        df = df[(j <= df.id) & (df.id <= k)]
    elif j:
        assert not (cohort is None), 'if j is specified, cohort must be specified'
        df = df[(df.id == j)]

    if paradigm:
        assert paradigm in df.paradigm.cat.categories, '{} is not a listed paradigm'.format(paradigm)
        df = df[(df.paradigm == paradigm)]

    return df[df.valid == True].copy()

def layout_dummy(df, key, template,
        strftime='%Y-%m-%d-%H%M'):
        df[key] = Series(
                data = [template.format(
                    r.cohort,
                    r.id,
                    r.paradigm,
                    r.date.strftime(strftime))
                    for r in df.itertuples()],
                index = df.index)

def layout_sdummy(df, key, template, urname, scale_type,
        strftime='%Y-%m-%d-%H%M'):
        df[key] = Series(
                data = [template.format(
                    r.cohort,
                    r.id,
                    r.paradigm,
                    r.date.strftime(strftime),
                    urname,
                    scale_type)
                    for r in df.itertuples()],
                index = df.index)



"""

This module contains methods for foreground detection.

"""

# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from skimage import filters

from pandas import DataFrame

def fit_foreground(data, ep):
    """
    Fit foreground

    Parameters
    ----------
    data : ndarray, shape (n,a,b,c)
    ep : int

    Returns
    -------
    DataFrame :
        Summary of thresholds, mean signals, and coefficients of
        variation.

    Notes
    -----
    This will overwrite your data and set all estimated background
    pixels to nan.
    """
    if np.isnan(data).any():
        print('  …data contain {} nan values, set to 0'.format(np.isnan(data).sum()))
        data[np.isnan(data)] = 0

    dat = np.moveaxis(data, ep+1, 1)
    cn = dat.shape[0]
    pn = dat.shape[1]

    ct = np.zeros(cn)
    pt = np.zeros((pn,cn))

    for c in range(cn):
        ct[c] = filters.threshold_otsu(dat[c])
        for p in range(pn):
            pt[p,c] = filters.threshold_otsu(dat[c,p])

    dat = np.moveaxis(dat, (0,1), (-1,-2))
    msk = (dat < ct) | (dat < pt)
    dat[msk] = np.nan
    return ct, pt



# Copyright 2016-2018 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Fit a signal model to FMRI data

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

########################################################################
# Input arguments
########################################################################

    input_files = parser.add_argument_group(
            """input files""",
            """Paths or templates for the paths to input fiels.""")

    input_files.add_argument('--session',
            default='../data/ses/{2}/{0}-{1:04d}-{2}-{3}.ses',
            help='input file;' + hp.session)

    # TODO: also give the option --reference-maps None or none,
    # which will assume no movement. This is useful for analysing
    # phantom data.
    input_files.add_argument('--reference-maps',
            default='../data/ref/{2}/{0}-{1:04d}-{2}-{3}.ref',
            help='input file;' + hp.reference_maps)

    input_files.add_argument('--population-map',
            default='../data/pop/{2}/{4}/{0}-{1:04d}-{2}-{3}-{4}.pop',
            help='input file;' + hp.population_map)

    input_files.add_argument('--population-space',
            default='reference',
            help=hp.population_space)

########################################################################
# Output arguments
########################################################################

    output_files = parser.add_argument_group(
            """output files""",
            """Paths or templates for the paths to output field.""")

    output_files.add_argument('--fit',
            default='../data/fit/{2}/{4}/{5}/{0}-{1:04d}-{2}-{3}-{4}-{5}.fit',
            help='output file;' + hp.sfit)

    output_files.add_argument('-o', '--protocol-log',
            default='logs/{}-fmrifit.pkl',
            help=hp.protocol_log)

########################################################################
# Arguments specific for the RSM Signal Model: design matrix
########################################################################

    signal_model = parser.add_argument_group(
            """define the signal model""",
            """Arguments which define the model that shall be fitted to
            the data by the RS estimator.""")

    signal_model.add_argument('--formula',
            default='C(task)/C(block, Sum)',
            help="""formula""")

    signal_model.add_argument('--parameter',
            default=['intercept', 'task'],
            nargs='+',
            help="""parameter""")

########################################################################
# Arguments specific for the RSM Signal Model: irritation II
########################################################################

    experimental_design = parser.add_argument_group(
            """define the experimental design""",
            """Arguments which define the experimental design of the
            session.""")

    experimental_design.add_argument('--stimulus-block',
            default='stimulus',
            help=hp.stimulus_block)

    experimental_design.add_argument('--control-block',
            default='control',
            help=hp.control_block)

    experimental_design.add_argument('--acquisition-burn-in',
            default=4,
            type=int,
            help=hp.burn_in)

    experimental_design.add_argument('--offset-beginning',
            type=float,
            default=5.242,
            help=hp.offset_beginning)

    experimental_design.add_argument('--offset-end',
            type=float,
            default=1.242,
            help=hp.offset_end)

########################################################################
# Arguments specific for the RSM Signal Model: weighting kernel
########################################################################

    weighting = parser.add_argument_group(
            """control the weighting of observations""",
            """Arguments which control the weighting scheme used in RS
            estimation.""")

    weighting.add_argument('--scale',
            type=float,
            help=hp.scale)

    weighting.add_argument('--scale-type',
            default='max',
            choices=['diagonal','max','min'],
            help=hp.scale_type)

    weighting.add_argument('--factor',
            type=float,
            default=3,
            help=hp.factor)

    weighting.add_argument('--mass',
            type=float,
            help=hp.mass)

########################################################################
# Arguments specific for the RSM Signal Model: where to fit
########################################################################

    where_to_fit = parser.add_argument_group(
            """where to fit""",
            """By default, fmrifit will respect the brain mask which is
            saved in the population map, and it will fit the model at
            all points within this mask. This default behaviour can, of
            course, be changed.""")

    where_to_fit.add_argument('--ignore-mask',
            action='store_true',
            help="""Ignore any brain masks saved in the respective
            population map.""")

    # TODO: mask: not implemented yet
    #where_to_fit.add_argument('--mask',
    #        action='store_true',
    #        help=hp.population_mask)

    where_to_fit.add_argument('--at-slice',
            type=int,
            nargs='+',
            help="""only fit the model at a slice of the index lattice""")

########################################################################
# Arguments specific for using the protocol API
########################################################################

    to_process = parser.add_argument_group(
            """specifying the protocol entries to process""",
            """Arguments which give control which protocol entries to
            process. If no protocol file is given, it will be checked
            if files being processed comply to the given information.""")

    to_process.add_argument('--protocol',
            help=hp.protocol)

    to_process.add_argument('--cohort',
            help=hp.cohort)

    to_process.add_argument('--id',
            type=int,
            nargs='+',
            help=hp.j)

    to_process.add_argument('--datetime',
            help=hp.datetime)

    to_process.add_argument('--paradigm',
            help=hp.paradigm)

    to_process.add_argument('--strftime',
            default='%Y-%m-%d-%H%M',
            help=hp.strftime)

########################################################################
# Miscellaneous
########################################################################

    handling = parser.add_argument_group(
            """file handling""")

    lock_handling = handling.add_mutually_exclusive_group()

    lock_handling.add_argument('-r', '--remove-lock',
            action='store_true',
            help=hp.remove_lock.format('fit'))

    lock_handling.add_argument('-i', '--ignore-lock',
            action='store_true',
            help=hp.ignore_lock.format('fit'))

    file_handling = handling.add_mutually_exclusive_group()

    file_handling.add_argument('-f', '--force',
            action='store_true',
            help=hp.force.format('fit'))

    file_handling.add_argument('-s', '--skip',
            action='store_true',
            help=hp.skip.format('fit'))

########################################################################
# Arguments
########################################################################

    detect = parser.add_argument_group(
            """detections""",
            """If these things have not already happend, you may perform
            these operations know.""")

    detect.add_argument('--grubbs',
            type=float,
            help=hp.grubbs)

    detect.add_argument('--detect-foreground',
            action='store_true',
            help=hp.detect_foreground)

########################################################################
# Multiprocessing
########################################################################

    misc = parser.add_argument_group(
            """miscellaneous""")

    misc.add_argument('-v', '--verbose',
            action='count',
            default=0,
            help=hp.verbose)

    misc.add_argument('-j', '--cores',
            type=int,
            help=hp.cores)

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ..df import get_df

from ...lock import Lock

from ...load import load_result, load_session, load_refmaps, load_population_map

from ...name import Identifier

from ...protocol import layout_dummy, layout_sdummy

from ...session import Session

from ...reference import ReferenceMaps

from ...smodel import SignalModel, Result

from ...pmap import PopulationMap, pmap_scanner, pmap_reference

import pandas as pd

import datetime

import sys

import os

from os.path import isfile, isdir, join

from multiprocessing.dummy import Pool as ThreadPool

import numpy as np

########################################################################

def call(args):
    output = args.protocol_log.format(
            datetime.datetime.now().strftime('%Y-%m-%d-%H%M'))

    if args.strftime == 'short':
        args.strftime = '%Y-%m-%d'

    ####################################################################
    # Parse protocol
    ####################################################################

    df = get_df(args, fall_back=args.session)

    if df is None:
        sys.exit()

    ####################################################################
    # Add file layout
    ####################################################################

    df_layout = df.copy()

    layout_dummy(df_layout, 'ses',
            template=args.session,
            strftime=args.strftime
            )

    layout_dummy(df_layout, 'ref',
            template=args.reference_maps,
            strftime=args.strftime
            )

    layout_sdummy(df_layout, 'pop',
            template=args.population_map,
            urname=args.population_space,
            scale_type=args.scale_type,
            strftime=args.strftime
            )

    layout_sdummy(df_layout, 'file',
            template=args.fit,
            urname=args.population_space,
            scale_type=args.scale_type,
            strftime=args.strftime
            )

    ####################################################################
    # Respect the mask mask
    ####################################################################

    mask = True

    if args.ignore_mask:
        mask = False

    ####################################################################
    # Fit at slice
    ####################################################################

    if args.at_slice:
        fit_at_slice = True
        slice_object = (slice(args.at_slice[0], args.at_slice[1]),
                        slice(args.at_slice[2], args.at_slice[3]),
                        slice(args.at_slice[4], args.at_slice[5]))
    else:
        fit_at_slice = False
        slice_object = None

    ####################################################################
    # Apply wrapper
    ####################################################################

    df['locked'] = False

    def wm(r):
        name = Identifier(cohort=r.cohort, j=r.id, datetime=r.date, paradigm=r.paradigm)

        try:
            dfile = os.path.dirname(r.file)
            if dfile and not isdir(dfile):
                os.makedirs(dfile)
        except Exception as e:
            print('{}: {}'.format(name.name(), e))

        wrapper(name              = name,
                df                = df,
                index             = r.Index,
                remove_lock       = args.remove_lock,
                ignore_lock       = args.ignore_lock,
                force             = args.force,
                skip              = args.skip,
                verbose           = args.verbose,
                file              = r.file,

                file_ses = r.ses,
                file_ref = r.ref,
                file_pop = r.pop,

                vb                   = args.population_space,
                stimulus_block       = args.stimulus_block,
                control_block        = args.control_block,
                scale_type           = args.scale_type,
                scale                = args.scale,
                factor               = args.factor,
                mass                 = args.mass,
                mask                 = mask,
                offset               = args.offset_beginning,
                preset               = args.offset_end,
                sgnf                 = args.grubbs,
                detect_foreground    = args.detect_foreground,
                burn_in              = args.acquisition_burn_in,
                formula              = args.formula,
                parameter            = args.parameter,

                fit_at_slice         = fit_at_slice,
                slice_object         = slice_object,
                )

    it =  df_layout.itertuples()

    if len(df_layout) > 1 and ((args.cores is None) or (args.cores > 1)):
        try:
            pool = ThreadPool(args.cores)
            results = pool.map(wm, it)
            pool.close()
            pool.join()
        except Exception as e:
            pool.close()
            pool.terminate()
            print('Pool execution has been terminated')
            print(e)
        finally:
            files = df_layout.ix[df.locked, 'file'].values
            if len(files) > 0:
                for f in files:
                    print('Unlock: {}'.format(f))
                    os.remove(f)
            del df['locked']
    else:
        try:
            print('Process protocol entries sequentially')
            for r in it:
                wm(r)
        finally:
            files = df_layout.ix[df.locked, 'file'].values
            if len(files) > 0:
                for f in files:
                    print('Unlock: {}'.format(f))
                    os.remove(f)
            del df['locked']

    ####################################################################
    # Write protocol
    ####################################################################

    if args.verbose:
        print('Save: {}'.format(output))

    dfile = os.path.dirname(output)
    if dfile and not isdir(dfile):
       os.makedirs(dfile)

    df.to_pickle(output)

#######################################################################

def wrapper(name, df, index, remove_lock, ignore_lock, force, skip,
        verbose, file, file_ses, file_ref, file_pop, vb,
        stimulus_block, control_block, scale_type, scale, factor, mass,
        mask,
        offset, preset, sgnf,
        detect_foreground, burn_in,
        formula, parameter,
        fit_at_slice, slice_object):

    if isfile(file):
        instance = load_result(file, name, df, index, vb, verbose)
        if type(instance) is Lock:
            if remove_lock or ignore_lock:
                if verbose:
                    print('{}: Remove lock'.format(name.name()))
                instance.unlock()
                if remove_lock:
                    return
            else:
                if verbose:
                    print('{}: Locked'.format(name.name()))
                return
        else:
            if df.ix[index,'valid'] and not force:
                if verbose:
                    print('{}: Valid'.format(name.name()))
                return
            else:
                if skip:
                    if verbose:
                        print('{}: Invalid'.format(name.name()))
                    return

    if skip:
        return

    if verbose:
        print('{}: Lock: {}'.format(name.name(), file))

    lock = Lock(name, 'fmrifit', file)
    df.ix[index, 'locked'] = True
    lock.save(file)
    df.ix[index,'valid'] = True

    ####################################################################
    # Load session instance from disk
    ####################################################################

    session = load_session(file_ses, name, df, index, verbose)
    if lock.conditional_unlock(df, index, verbose):
        return

    if detect_foreground:
        if verbose:
            print('{}: Detect foreground'.format(name.name()))
        session.fit_foreground()

    ####################################################################
    # Load reference maps from disk
    ####################################################################

    reference_maps = load_refmaps(file_ref, name, df, index, verbose)
    if lock.conditional_unlock(df, index, verbose):
        return

    if (sgnf is not None) and (not np.isclose(sgnf, 1)):
        if verbose:
            print('{}: Detect outlying scans'.format(name.name()))
        reference_maps.detect_outlying_scans(sgnf)

    ####################################################################
    # Load or created population map from disk (as needed)
    ####################################################################

    population_map = load_population_map(file_pop, name, df, index, vb, verbose)
    if lock.conditional_unlock(df, index, verbose):
        return

    if verbose:
        print('{}: Population space is: {}'.format(name.name(),
            population_map.diffeomorphism.vb))

    ########################################################################
    # Create signal model instance
    ########################################################################

    if verbose:
        print('{}: Configure signal model'.format(name.name()))

    smodel = SignalModel(
        session=session,
        reference_maps=reference_maps,
        population_map=population_map)

    smodel.create_irritation_design(
            s=stimulus_block,
            c=control_block,
            offset=offset,
            preset=preset)

    smodel.create_data_matrix(burn_in=burn_in, verbose=verbose)

    smodel.set_hyperparameters(
            scale_type=scale_type,
            scale=scale,
            factor=factor,
            mass=mass)

    if verbose > 1:
        print(smodel.describe())

    if verbose:
        print('{}: Create design matrix'.format(name.name()))

    smodel.create_design_matrix(
            formula=formula,
            parameter=parameter,
            verbose=verbose)

    if fit_at_slice:
        if verbose:
            print('{}: Start fit at {}'.format(name.name(), slice_object))

        coordinates = smodel.population_map.diffeomorphism.coordinates()
        coordinates = coordinates[slice_object]

        result = smodel.fit_at_subject_coordinates(coordinates,
                mask=mask, verbose=verbose)

    else:
        result = smodel.fit(mask=mask, verbose=verbose)

    if verbose:
        print('{}: Save: {}'.format(name.name(), file))

    try:
        result.save(file)
        df.ix[index,'locked'] = False
    except Exception as e:
        df.ix[index,'valid'] = False
        print('{}: Unable to write: {}'.format(name.name(), file))
        print('{}: Exception: {}'.format(name.name(), e))
        lock.conditional_unlock(df, index, verbose, True)

    if verbose > 1:
        print('{}: Description'.format(name.name()))
        print(result.describe())



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

The circle constant

"""

from numpy import pi

tau = 2*pi



# Copyright 2016-2018 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Command line tool to create a sample of fits of the FMRI signal model

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

########################################################################
# Input arguments
########################################################################

# TODO: make population space optional, save ati in Sample and also the
# cfactor field!!

    parser.add_argument('sample',
            help="""path where to save the sample""")

    parser.add_argument('population_space',
            help="""path to a population space""")

    parser.add_argument('population_ati',
            help="""path to a population ATI reference""")

    parser.add_argument('--fit',
            default='../data/fit/{2}/{4}/{5}/{0}-{1:04d}-{2}-{3}-{4}-{5}.fit',
            help='output file;' + hp.sfit)

    parser.add_argument('--scale-type',
            default='max',
            choices=['diagonal','max','min'],
            help=hp.scale_type)

########################################################################
# Output arguments
########################################################################

    parser.add_argument('-o', '--protocol-log',
            default='logs/{}-fmrisample.pkl',
            help=hp.protocol_log)

########################################################################
# Arguments specific for using the protocol API
########################################################################

    parser.add_argument('protocol',
            help=hp.protocol)

    parser.add_argument('--cohort',
            help=hp.cohort)

    parser.add_argument('--id',
            type=int,
            nargs='+',
            help=hp.j)

    parser.add_argument('--paradigm',
            help=hp.paradigm)

    parser.add_argument('--strftime',
            default='%Y-%m-%d-%H%M',
            help=hp.strftime)

########################################################################
# Miscellaneous
########################################################################

    parser.add_argument('-f', '--force',
            action='store_true',
            help=hp.force.format('result'))

    parser.add_argument('-v', '--verbose',
            action='store_true',
            help=hp.verbose)

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ..df import get_df

from ...smodel import Result

from ...pmap import PopulationMap

from ...sample import Sample

from ...load import load, load_result

from ...name import Identifier

from ...protocol import layout_sdummy

import numpy as np

import pandas as pd

import datetime

import sys

import os

from os.path import isfile, isdir, join

########################################################################

def call(args):

    try:
        population_space = load(args.population_space)
    except Exception as e:
        print('Unable to read: {}'.format(args.population_space))
        print('Exception: {}'.format(e))
        exit()

    try:
        population_ati = load(args.population_ati)
    except Exception as e:
        print('Unable to read: {}'.format(args.population_space))
        print('Exception: {}'.format(e))
        exit()

    ####################################################################

    output = args.protocol_log.format(
            datetime.datetime.now().strftime('%Y-%m-%d-%H%M'))

    if args.strftime == 'short':
        args.strftime = '%Y-%m-%d'

    ####################################################################
    # Parse protocol
    ####################################################################

    df = get_df(args)

    if df is None:
        sys.exit()

    ####################################################################
    # Add file layout
    ####################################################################

    df.reset_index(inplace=True, drop=True)

    df_layout = df.copy()

    layout_sdummy(df_layout, 'file',
            template=args.fit,
            urname=population_space.name,
            scale_type=args.scale_type,
            strftime=args.strftime
            )

    ####################################################################
    # Apply wrapper
    ####################################################################

    if not isfile(args.sample) or args.force:
        if args.verbose:
            print('Create population sample…'.format(args.sample))

        statistics = np.empty(population_space.shape + (2,len(df),))
        statistics [ ... ] = np.nan

        for r in df_layout.itertuples():
            name = Identifier(cohort=r.cohort, j=r.id, datetime=r.date, paradigm=r.paradigm)

            wrapper(
                    name             = name,
                    df               = df,
                    index            = r.Index,
                    file             = r.file,
                    population_space = population_space,
                    population_ati   = population_ati,
                    statistics       = statistics,
                    verbose          = args.verbose,
                    )

        sample = Sample(
                population_space = population_space,
                covariates = df,
                statistics = statistics)

        sample = sample.filter()

        if args.verbose:
            print('Save: {}'.format(args.sample))

        sample.save(args.sample)

    else:
        if args.verbose:
            print('Parse: {}'.format(args.sample))
        sample = load(args.sample)

    if args.verbose:
        print('Description of the population space:')
        print(sample.population_space.describe())
        print('Description of the sample:')
        print(sample.describe())

    ####################################################################
    # Write protocol
    ####################################################################

    if args.verbose:
        print('Save: {}'.format(output))

    dfile = os.path.dirname(output)
    if dfile and not isdir(dfile):
       os.makedirs(dfile)

    df.to_pickle(output)

########################################################################

def wrapper(name, df, index, file, population_space, population_ati,
        statistics, verbose):

    result = load_result(file, name, df, index, population_space.name, verbose)

    if not df.ix[index,'valid']:
        return

    intercept = result.get_field('intercept','point')
    c = population_ati.data / intercept.data

    beta = result.get_field('task','point')
    beta_stderr = result.get_field('task','stderr')

    statistics[...,0,index] = c*beta.data
    statistics[...,1,index] = (c*beta_stderr.data)**2

    #statistics[...,2,index] = c



# Copyright 2018 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Creates a scalar unit field for ATI reference by smoothing the input
image to given curvature.

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

    parser.add_argument('scale',
            type=float,
            help="""scale (standard deviation) of the kernel""")

    parser.add_argument('input',
            help="""input image file""")

    parser.add_argument('output',
            help="""output image file""")

    parser.add_argument('--ignore-mask',
            action='store_true',
            help="""ignore image mask""")

    parser.add_argument('--name',
            help="""if not given, the name will be set equal to the
            input image.""")

    parser.add_argument('--factor',
            type=float,
            default=3,
            help="""factor""")

    parser.add_argument('-v', '--verbose',
            action='count',
            default=0,
            help=hp.verbose)

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ... import load

from ...diffeomorphisms import Image

from ...fit import fit_field, extract_field

import numpy as np

import time

########################################################################

def call(args):
    img = load(args.input)
    assert type(img) is Image

    if args.verbose > 1:
        print(img.describe())

    if args.name:
        name = args.name
    else:
        name = img.name

    if not args.ignore_mask:
        img = img.mask(img.get_mask())

    coordinates = img.coordinates()
    coordinates.shape

    if args.ignore_mask:
        mask = None
    else:
        mask          = img.get_mask()

    endog         = img.data.ravel()
    exog          = np.ones_like(endog).reshape((-1,1))
    agc           = coordinates.reshape((-1,3))

    old_settings = np.seterr(divide='raise', invalid='raise')
    time0 = time.time()

    statistics, parameter_dict, value_dict = fit_field(
            coordinates   = coordinates,
            mask          = mask,
            endog         = endog,
            exog          = exog,
            agc           = agc,
            scale         = args.scale,
            radius        = args.factor*args.scale,
            verbose       = args.verbose,
            name          = img.name)

    time1 = time.time()
    np.seterr(**old_settings)

    field = extract_field(
            field=statistics,
            param='intercept',
            value='point',
            parameter_dict={'intercept':0},
            value_dict=value_dict)

    result = Image(reference=img.reference, data=field, name=name)
    result.save(args.output)

    if args.verbose:
        time_spend = time1 - time0
        print('{}: Time needed for the fit: {:.2f} min'.format(
            result.name, time_spend / 60))
        print('{}: Time needed for the fit: {:.2f} h'  .format(
            result.name, time_spend / 60**2))

    # TODO: test result.mask(img.to_mask()) should have no effect
    # mytest = np.isfinite(result.statistics).all(axis=(-1,-2))



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Defines a class for the FMRI population model and its fits.

"""

from .sample import Sample

from .meta import fit_field

from patsy import dmatrix

import numpy as np

import pickle

class PopulationModel:
    """
    The FMRI population model

    Parameters
    ----------
    sample : Sample
    formula_like : str
        A formula like object that is understood by patsy.
    exog : ndarray
        If None, will be created from formula_like
        Directly specifying a design matrix, i.e., providing exog
        will take presendence from the formula interface.
    mask : None, bool, ndarray, or str
        If False or None, the population model will be fitted only
        at points, at which the population model is identifiable.
        If True or 'template', the population model will,
        additionally, only be fitted at points, at which the
        template in the population has valid intensities (i.e. > 0
        and not NAN). If 'sample', the population model will be
        fitted only at points at which the sample provides valid
        summary statistics for *all* fields in the sample.
    """

    def __init__(self, sample, formula_like=None, exog=None, mask=True):
        assert type(sample) is Sample, 'sample must be of type Sample'

        self.covariates = sample.covariates
        self.statistics = sample.statistics
        self.population_space = sample.population_space

        if (formula_like is not None) and (exog is None):
            exog = np.asarray(dmatrix(
                formula_like=formula_like,
                data=self.covariates))

        if mask is True:
            mask = 'template'

        if type(mask) is str:
            if mask == 'template':
                mask = self.population_space.get_mask()
            elif mask == 'sample':
                mask = np.isfinite(self.statistics).all(axis=(-1,-2))

        assert mask.any(), 'mask is empty, i.e., there exists no valid pixels in this mask'

        self.formula_like = formula_like
        self.exog = exog
        self.mask = mask

    def fit(self):
        """
        Fit the population model to data
        """
        statistics = fit_field(
                obs=self.statistics,
                design=self.exog,
                mask=self.mask)

        return MetaResult(
                statistics=statistics,
                formula_like=self.formula_like,
                exog=self.exog,
                population_space=self.population_space)

    ####################################################################
    # Save instance to and from disk
    ####################################################################

    def save(self, file, **kwargs):
        """
        Save model instance to disk

        This will save the current model instance to disk for later use.

        Parameters
        ----------
        file : str
            File name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)

class MetaResult:
    def __init__(self, statistics, formula_like, exog, population_space):
        self.statistics = statistics
        self.formula = formula_like
        self.exog = exog
        self.population_space = population_space
        self.reference = population_space.reference

    def get_activation_patterns(self):
        return np.moveaxis(self.statistics[...,0,:-1], -1, 0)

    def get_tvalues(self):
        return np.moveaxis(self.statistics[...,1,:-1], -1, 0)

    def get_heterogeneities(self):
        return self.statistics[...,0,-1]

    def get_degrees_of_freedom(self):
        return self.statistics[...,1,-1]

    ####################################################################
    # Save instance to and from disk
    ####################################################################

    def save(self, file, **kwargs):
        """
        Save model instance to disk

        This will save the current model instance to disk for later use.

        Parameters
        ----------
        file : str
            File name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Query protocol files

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

########################################################################
# Input arguments
########################################################################

    parser.add_argument('protocol',
            help=hp.protocol)

########################################################################
# Output arguments
########################################################################

    parser.add_argument('-o', '--output',
            help="""output file""")

########################################################################
# Arguments specific for using the protocol API
########################################################################

    to_process = parser.add_argument_group(
            """specifying the protocol entries to process""",
            """Arguments which give control which protocol entries to
            process.""")

    to_process.add_argument('--cohort',
            help=hp.cohort)

    to_process.add_argument('--id',
            type=int,
            nargs='+',
            help=hp.j)

    to_process.add_argument('--datetime',
            help=hp.datetime)

    to_process.add_argument('--paradigm',
            help=hp.paradigm)

    to_process.add_argument('--strftime',
            default='%Y-%m-%d-%H%M',
            help=hp.strftime)

########################################################################
# Arguments specific for using fmriprotocol
########################################################################

    parser.add_argument('--update',
            nargs='+',
            help="""update field 'valid' in the protocol file with the
            entries in the file UPDATE. This step will
            be applied *after* applying your filters.""")

    parser.add_argument('--query',
            help="""apply a QUERY to the data frame.""")

    parser.add_argument('--head',
            type=int,
            help="""print the first HEAD number of lines to standard
            output.""")

    parser.add_argument('--to-csv',
            help="""export protocol file to CSV""")

########################################################################
# Miscellaneous
########################################################################

    misc = parser.add_argument_group(
            """miscellaneous""")

    misc.add_argument('-f', '--force',
            action='store_true',
            help=hp.force.format('protocol'))

    misc.add_argument('-v', '--verbose',
            action='count',
            default=0,
            help=hp.verbose)

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ..df import filter_df

from ...name import Identifier

import pandas as pd

import sys

import os

from os.path import isfile, isdir, join

########################################################################

def call(args):
    if args.strftime == 'short':
        args.strftime = '%Y-%m-%d'

    try:
        df = pd.read_pickle(args.protocol)
        if args.verbose:
            print('Read: {}'.format(args.protocol))
    except Exception as e:
        print('Unable to read protocol file {}'.format(args.protocol))
        print('Exception: {}'.format(e))
        sys.exit()

    if args.verbose:
        valid = df.groupby(['cohort','paradigm','valid']).epi.agg(['count'])
        print('Number of entries in the list: {}'.format(len(df)))
        if args.verbose > 1:
            print()
            print(valid)
            print()

    df = filter_df(df, args.cohort, args.id, args.paradigm)

    if args.verbose:
        valid = df.groupby(['cohort','paradigm','valid']).epi.agg(['count'])
        print('Number of relevant entries in the list: {}'.format(len(df)))
        if args.verbose > 1:
            print()
            print(valid)
            print()

    if args.update:
        for upfile in args.update:
            up = pd.read_pickle(upfile)

            if 'date' in up.index.names:
                on = ['cohort','id','paradigm','date']
            elif 'paradigm' in up.index.names:
                on = ['cohort','id','paradigm']
            else:
                on = ['cohort','id']

            df = (df.join(up, on=on, lsuffix='_')
                    .assign(valid=lambda x: x.valid.fillna(True))
                    .assign(valid=lambda x: x.valid & x.valid_)
                    .drop('valid_', axis=1))

        if args.verbose:
            valid = df.groupby(['cohort','paradigm','valid']).epi.agg(['count'])
            print('Number of entries in the list after --update: {}'.format(len(df)))
            if args.verbose > 1:
                print()
                print(valid)
                print()
            if args.verbose > 2:
                print(df.head(args.head))
                print()

    if args.query:
        df = df.query(args.query).copy()

        if args.verbose:
            valid = df.groupby(['cohort','paradigm','valid']).epi.agg(['count'])
            print('Number of entries in the list after --query: {}'.format(len(df)))
            if args.verbose > 1:
                print()
                print(valid)
                print()

    df = filter_df(df, args.cohort, args.id, args.paradigm)

    if df is None:
        print('Number of relevant entries in the list: {}'.format(len(df)))
    else:
        if args.verbose:
            valid = df.groupby(['cohort','paradigm','valid']).epi.agg(['count'])
            print('Number of relevant entries in the list: {}'.format(len(df)))
            if args.verbose > 1:
                print()
                print(valid)
                print()

    if df is None:
        sys.exit()

    df = filter_df(df, args.cohort, args.id, args.paradigm)

    if args.head:
        if args.verbose:
            print()
            print("""First {:d} valid entries in the final protocol:
---------------------------------------------""".format(args.head))
        print(df.head(args.head))
        print()

    if args.output:
        if not isfile(args.output) or args.force:
            if args.verbose:
                print('Save: {}'.format(args.output))

            df.to_pickle(args.output)
        else:
            print('File already exists: {}, use --force to overwrite'.format(args.output))

    if args.to_csv:
        if not isfile(args.to_csv) or args.force:
            if args.verbose:
                print('Save: {}'.format(args.to_csv))
            df.to_csv(args.to_csv, index=False)
        else:
            print('File already exists: {}, use --force to overwrite'.format(args.to_csv))



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Defines classes for diffeomorphisms which are not necessarily affine
transformations.

"""

from .name import Identifier

from .affines import Affine

import pickle

import numpy as np

import numpy.ma as ma

from numpy.linalg import inv

from scipy.ndimage import map_coordinates

class Diffeomorphism:
    """
    A diffeomorphism ψ from standard space to subject space

    Parameters
    ----------
    reference : Affine or ndarray, shape (4,4)
        An affine transformation
    shape : tuple
        Shape of template data in standard space.
    vb : str or Identifier
        An object that will be used to identify the domain (vb) of this
        diffeomorphism.
    nb : str or Identifier
        An object that will be used to identify the image (nb) of this
        diffeomorphism.
    metadata : dict
        If meta data is provided which give information on which data
        had been used for the fit: which had provided, should be a dict
        with at least the following fields: ``{'vb_file':
        path/which/defined/the/domain, 'nb_file':
        path/which/defined/the/image,}``.

    Notes
    -----
    A reasonable subclass of Diffeomorphism must define the attributes
    :func:`apply_to_index`, :func:`apply_to_indices`, and :func:`apply`.
    """
    def __init__(self, reference, shape, vb=None, nb=None, metadata=None):
        self.reference = reference
        self.shape = shape
        self.vb = vb
        self.nb = nb
        self.metadata = metadata

    def apply_to_index(self, index):
        """
        Apply diffeomorphism to the point with given index

        Parameters
        ----------
        index : tuple
            Should be a tuple of integers

        Returns
        -------
        ndarray
        """
        pass

    def apply_to_indices(self, indices):
        """
        Apply diffeomorphism to the points with given indices

        Parameters
        ----------
        index : tuple
            Should be a tuple of slices or ellipses

        Returns
        -------
        ndarray
        """
        pass

    def apply(self, coordinates):
        """
        Apply diffeomorphism to the point at given coordinate

        This is the identity; it will return the same coordinates

        Parameters
        ----------
        index : tuple
            Should be a tuple of integers

        Returns
        -------
        ndarray
        """
        pass

    def coordinates(self):
        """
        Coordinates in the image

        Returns
        -------
        ndarray
        """
        x,y,z = self.shape
        indices = ((slice(0,x), slice(0,y), slice(0,z)))
        return self.apply_to_indices(indices)

    def coordinates_domain(self):
        """
        Coordinates in the domain

        Returns
        -------
        ndarray
        """
        x,y,z = self.shape
        indices = ((slice(0,x), slice(0,y), slice(0,z)))
        return self.reference.apply_to_indices(indices)

    def describe(self):
        if type(self.nb) is Identifier:
            description = """
            Population space
            ----------------
            Name:  {}
            Shape: {}
            Volume of a voxel: {:.2f}

            Subject reference space
            ----------------
            Cohort:   {}
            Subject:  {}
            Paradigm: {}
            """
            return description.format(
                    self.vb,
                    self.shape,
                    self.reference.volume(),
                    self.nb.cohort,
                    self.nb.j,
                    self.nb.paradigm,
                    )
        else:
            description = """
            Population space
            ----------------
            Name:  {}
            Shape: {}
            Volume of a voxel: {:.2f}

            Subject reference space
            ----------------
            name:   {}
            """
            return description.format(
                    self.vb,
                    self.shape,
                    self.reference.volume(),
                    self.nb
                    )

    def save(self, file, **kwargs):
        """
        Save instance to disk

        Parameters
        ----------
        file : str
            A file name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)

class Identity(Diffeomorphism):
    """
    The identity diffeomorphism ψ.
    """
    def __init__(self, reference, shape, vb=None, nb=None):
        self.reference = reference
        self.shape = shape
        self.vb = vb
        self.nb = nb

    def apply_to_index(self, index):
        """
        Apply diffeomorphism to the point with given index

        Parameters
        ----------
        index : tuple
            Should be a tuple of integers.

        Returns
        -------
        ndarray
        """
        return self.reference.apply_to_index(index)

    def apply_to_indices(self, indices):
        """
        Apply diffeomorphism to the points with given indices

        Parameters
        ----------
        index : tuple
            Should be a tuple of slices or ellipses.

        Returns
        -------
        ndarray
        """
        return self.reference.apply_to_indices(indices)

    def apply(self, coordinates):
        """
        Apply diffeomorphism to the point at given coordinate

        This is the identity; it will return the same coordinates

        Parameters
        ----------
        index : tuple
            Should be a tuple of integers.

        Returns
        -------
        ndarray
        """
        return coordinates

# TODO: why not treat an Image as a masked array with values 0 and nan.

class Image(Identity):
    """
    Holds all information of a 3D-image.

    Parameters
    ----------
    reference : Affine or ndarray, shape (4,4)
        An affine transformation that maps indices in the data
        array to coordinates in Euclidean space.
    data : np.ndarray
        The data array.
    name : str or Identifier
        An identifier or name for this image.
    """
    def __init__(self, reference, data, name=None):
        if type(reference) is Affine:
            self.reference = reference
        else:
            self.reference = Affine(reference)

        self.data = data
        self.shape = self.data.shape

        if name:
            self.name = name
            self.vb = name
            self.nb = name
        else:
            self.name = None
            self.vb = None
            self.nb = None

    def flatten(self, epi_code, xpic, ypic, slices=None):
        """
        Flatten a 3D image into a 2D image

        Parameters
        ----------
        epi_code : int

        Returns
        -------
        ndarray

        Notes
        -----
        Flattens by slicing the 3D image along the axis of the respected
        EPI code, and stack the slices into a xpic x ypic matrix
        """
        assert epi_code <=  3, 'epi code must be ≤ 3'
        assert epi_code >= -3, 'epi code must be ≥-3'
        assert epi_code !=  0, 'epi code must be different from 0'

        npic = xpic * ypic
        data = np.moveaxis(self.data,abs(epi_code)-1,0)
        filled, = np.where(~np.isnan(data).all(axis=(1,2)))
        if slices is None:
            select = np.linspace(0,len(filled)-1,npic,dtype='int')
            slices = filled[select]
        else:
            assert len(slices) == npic, 'number of slices must equal xpic*ypic'

        data = data[slices]

        imatrix = np.vstack(
                [np.hstack([data[i].T for i in range(j,xpic+j)])
                    for j in range(0,npic,xpic)])

        xticks = [i*data.shape[2] for i in range(xpic)]
        yticks = [i*data.shape[1] for i in range(ypic)]
        return imatrix, xticks, yticks, slices

    def mask(self, mask):
        """
        Mask the data array by mask

        Parameters
        ----------
        mask : ndarray, dtype(bool)
            An array of bool indicating the foreground

        Returns
        -------
        Image
            The same image but with data set to nan outside the mask
        """
        data = self.data.astype(float)
        data [~mask] = np.nan
        return Image(reference=self.reference, data=data, name=self.name)

    def get_mask(self, **kwargs):
        """
        Create a mask

        Returns
        -------
        ndarray, dtype(bool)
            boolean array which is *True* whenever values in the data are
            *valid*.
        """
        return ~( (np.isnan(self.data)) | np.isclose(self.data, 0, **kwargs) )

    def round(self):
        """
        Round data in the image to integer

        Returns
        -------
        Image
        """
        data = self.data.copy()
        data [ ~self.get_mask() ] = 0
        data = data.round().astype(int)
        return Image(reference=self.reference, data=data, name=self.name)

    def mean(self):
        """
        Calculates the mean signal in the image

        Notes
        -----
        Values of 0 or nan will not contribute to the mean signal.  If
        your image ``img`` contains 0 values, then ``im.data.mean()``
        and ``im.mean()`` will differ and this is intended.

        Returns
        -------
        float
        """
        data = ma.array(self.data, mask=~self.get_mask(), fill_value=0)
        return data.mean()

    def volume(self):
        """
        Calculates the volume occupied by the image

        Returns
        -------
        float
        """
        return self.get_mask().sum() * self.reference.volume()

    def describe(self):
        """
        Describe the image

        Returns
        -------
        str
        """
        description = """
        Shape: {}
        Volume of one pixel: {:.2f}
        Volume of image: {:.2f}
        """
        return description.format(
                self.shape,
                self.reference.volume(),
                self.volume())

    def save(self, file, **kwargs):
        """
        Save instance to disk

        Parameters
        ----------
        file : str
            A file name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)

class AffineTransformation(Diffeomorphism):
    """
    An affine transformation ψ mapping from a population space :math:`M`
    to a subject space :math:`R`. It is

    .. math::

        ψ (x) = A⋅x.

    for :math:`x∈M` and :math:`A` an affine transformation.

    A point [i,j,k] in the index space of the template of :math:`M` maps
    to the coordinates

    .. math::

        ψ(reference[i,j,k]) = A ⋅ reference[i,j,k]

    in the subject space.

    Parameters
    ----------
    reference : Affine or ndarray, shape (4,4)
        An affine transformation
    affine : Affine or ndarray, shape (4,4)
        An affine transformation
    """

    def __init__(self, reference, affine, shape, vb=None, nb=None):
        if type(reference) is Affine:
            self.reference = reference
        else:
            self.reference = Affine(reference)

        if type(affine) is Affine:
            self.affine = affine
        else:
            self.affine = Affine(affine)

        self.shape = shape
        self.vb = vb
        self.nb = nb

    def apply_to_index(self, index):
        return self.affine.dot(self.reference).apply_to_index(index)

    def apply_to_indices(self, indices):
        return self.affine.dot(self.reference).apply_to_indices(indices)

    def apply(self, coordinates):
        """
        Apply diffeomorphism to the point at given coordinate

        Parameters
        ----------
        index : tuple
            Should be a tuple of integers

        Returns
        -------
        ndarray
        """
        return self.affine.apply(coordinates)

class Warp(Diffeomorphism):
    """
    A diffeomorphism ψ mapping from a population space :math:`M` to a
    subject space $R$. Here, it is

    .. math::

        ψ (x) = δ(x).

    for a warp field :math:`δ`.

    A point [i,j,k] in the index space of the template of :math:`M` maps
    to the coordinates

    .. math::

        ψ(reference[i,j,k]) = warp[i,j,k]

    An example: The FSL FNIRT warp coefficients file is saved as a Warp
    instance.

    Parameters
    ----------
    reference : Affine or ndarray, shape (4,4)
        An affine transformation
    warp : ndarray (...,3)
        A warp field
    vb : str or Identifier
        An object that will be used to identify the domain (vb) of this
        diffeomorphism.
    nb : str or Identifier
        An object that will be used to identify the image (nb) of this
        diffeomorphism.
    metadata : dict
        If meta data is provided which give information on which data
        had been used for the fit: which had provided, should be a dict
        with at least the following fields: ``{'vb_file':
        path/which/defined/the/domain, 'nb_file':
        path/which/defined/the/image,}``.
    """
    def __init__(self, reference, warp, vb=None, nb=None, metadata=None):
        assert type(warp) is np.ndarray, 'warp must be numpy.ndarray'
        assert 3 == warp.shape[-1], 'last dimension of warp must be 3'

        if type(reference) is Affine:
            self.reference = reference
        else:
            self.reference = Affine(reference)

        self.warp = warp
        self.shape = self.warp.shape[:-1]
        self.vb = vb
        self.nb = nb

        if metadata is not None:
            self.metadata = metadata

    def apply_to_index(self, index):
        return self.warp[index]

    def apply_to_indices(self, indices):
        return self.warp[indices]

    def apply(self, coordinates):
        """
        Apply diffeomorphism to the point at given coordinate

        Parameters
        ----------
        index : tuple
            Should be a tuple of integers

        Returns
        -------
        ndarray

        Notes
        -----
        There are interpolations at place here, which makes this
        potentially slow for large queries.
        """
        indices = self.reference.inv().apply(coordinates)
        newx = map_coordinates(self.warp[...,0], indices.T.reshape(3,-1))
        newy = map_coordinates(self.warp[...,1], indices.T.reshape(3,-1))
        newz = map_coordinates(self.warp[...,2], indices.T.reshape(3,-1))
        return np.vstack((newx, newy, newz)).T

    def describe(self):
        if hasattr(self, 'metadata'):
            description = """
            Population space
            ----------------
            Name:  {}
            Shape: {}
            Volume of a voxel: {:.2f}

            Subject reference space
            ----------------
            Cohort:   {}
            Subject:  {}
            Paradigm: {}

            Meta data of the fit:
            ---------------------
            vb_file : {}
            vb_mask : {}
            nb_file : {}
            """

            try:
                vb_mask=self.metadata['vb_mask']
            except:
                vb_mask=None

            return description.format(
                    self.vb,
                    self.shape,
                    self.reference.volume(),
                    self.nb.cohort,
                    self.nb.j,
                    self.nb.paradigm,
                    self.metadata['vb_file'],
                    vb_mask,
                    self.metadata['nb_file'],
                    )
        else:
            description = """
            Population space
            ----------------
            Name:  {}
            Shape: {}
            Volume of a voxel: {:.2f}

            Subject reference space
            ----------------
            Cohort:   {}
            Subject:  {}
            Paradigm: {}
            """
            return description.format(
                    self.vb,
                    self.shape,
                    self.reference.volume(),
                    self.nb.cohort,
                    self.nb.j,
                    self.nb.paradigm,
                    )

class Displacement(Diffeomorphism):
    """
    A diffeomorphism ψ mapping from a population space :math:`M` to a
    subject space $R$. Here, it is

    .. math::

        ψ (x) = x + δ(x).

    for a displacement field :math:`δ`.

    A point [i,j,k] in the index space of the template of :math:`M` maps
    to the coordinates

    .. math::

        ψ(reference[i,j,k]) = reference[i,j,k] + displacement[i,j,k]

    Parameters
    ----------
    reference : Affine or ndarray, shape (4,4)
        An affine transformation.
    displacement : ndarray (?,?,?,3)
        A displacement field
    vb : str or Identifier
        An object that will be used to identify the domain (vb) of this
        diffeomorphism.
    nb : str or Identifier
        An object that will be used to identify the image (nb) of this
        diffeomorphism.
    metadata : dict
        If meta data is provided which give information on which data
        had been used for the fit: which had provided, should be a dict
        with at least the following fields: ``{'vb_file':
        path/which/defined/the/domain, 'nb_file':
        path/which/defined/the/image,}``.
    """

    def __init__(self, reference, displacement, vb=None, nb=None):
        assert type(displacement) is np.ndarray, 'displacement must be numpy.ndarray'
        assert 3 == displacement.shape[-1], 'last dimension of displacement must be 3'

        if type(reference) is Affine:
            self.reference = reference
        else:
            self.reference = Affine(reference)

        self.displacement = displacement
        self.shape = self.displacement.shape[:-1]
        self.vb = vb
        self.nb = nb

        if metadata is not None:
            self.metadata = metadata

    def apply_to_index(self, index):
        return self.reference.apply_to_index(index) + self.displacement[index]

    def apply_to_indices(self, indices):
        return self.reference.apply_to_indices(indices) + self.displacement[indices]

    # TODO: test
    def apply(self, coordinates):
        """
        Apply diffeomorphism to the point at given coordinate

        Parameters
        ----------
        index : tuple
            Should be a tuple of integers

        Returns
        -------
        ndarray

        Notes
        -----
        There are interpolations at place here, which makes this
        potentially slow for large queries.
        """
        indices = self.reference.inv().apply(coordinates)
        newx = map_coordinates(self.displacement[...,0], indices.T.reshape(3,-1))
        newy = map_coordinates(self.displacement[...,1], indices.T.reshape(3,-1))
        newz = map_coordinates(self.displacement[...,2], indices.T.reshape(3,-1))
        return coordinates + np.vstack((newx, newy, newz)).T

    def describe(self):
        if hasattr(self, 'metadata'):
            description = """
            Population space
            ----------------
            Name:  {}
            Shape: {}
            Volume of a voxel: {:.2f}

            Subject reference space
            ----------------
            Cohort:   {}
            Subject:  {}
            Paradigm: {}

            Meta data of the fit:
            ---------------------
            vb_file : {}
            vb_mask : {}
            nb_file : {}
            """

            try:
                vb_mask=self.metadata['vb_mask']
            except:
                vb_mask=None

            return description.format(
                    self.vb,
                    self.shape,
                    self.reference.volume(),
                    self.nb.cohort,
                    self.nb.j,
                    self.nb.paradigm,
                    self.metadata['vb_file'],
                    vb_mask,
                    self.metadata['nb_file'],
                    )
        else:
            description = """
            Population space
            ----------------
            Name:  {}
            Shape: {}
            Volume of a voxel: {:.2f}

            Subject reference space
            ----------------
            Cohort:   {}
            Subject:  {}
            Paradigm: {}
            """
            return description.format(
                    self.vb,
                    self.shape,
                    self.reference.volume(),
                    self.nb.cohort,
                    self.nb.j,
                    self.nb.paradigm,
                    )
